'''Test the catalog shelf module.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, Iterable, Set, Tuple, Union
import pytest

from .catalog import Catalog, CatalogLookupError
from .memory_catalog import MemoryCatalog
from .catalog_shelf import CatalogShelf
from .catalog_element_mixin import CatalogElementMixin

# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code


class Widget():
    '''This is the thing that is stored in WidgetCatalog.'''


class NamedWidget(Widget, CatalogElementMixin):
    '''Widget that properly inherits from CatalogElementMixin and thus has a name'''


class WidgetCatalog(MemoryCatalog[Widget]):
    '''This is the class we will exercise.'''


class WidgetNonMemoryCatalog(Catalog[Widget]):
    '''This is a non-memory catalog.'''

    @property
    def tagtypes(self) -> Set[str]:
        return set()

    def all_objects(self) -> Iterable[Widget]:
        return []

    def items(self) -> Iterable[Tuple[str, Widget]]:
        return []

    def tagvals(self, tagtype: str) -> Set[str]:
        return set()

    def lookup_by_name(self, name: str) -> Widget:
        raise CatalogLookupError()

    def lookup_by_tag_and(self, **tags: Union[str, Iterable[str]]) -> Dict[str, Widget]:
        return {}


def test_items():
    '''Register a widget into a shelf with an empty catalog at the start.'''
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1 = NamedWidget(name='widget1')
    catalog2.register(widget1)
    widget2 = NamedWidget(name='widget2')
    dut = CatalogShelf(catalog1, catalog2)
    dut.register(widget2)
    assert dict(dut.items()) == {
        'widget1': widget1,
        'widget2': widget2,
    }


def test_one_catalog():
    '''Load a catalog into the Shelf'''
    catalog1 = WidgetCatalog()
    widget1 = NamedWidget(name='widget1')
    catalog1.register(widget1)
    dut = CatalogShelf(catalog1)
    assert dut.lookup_by_name('widget1') == widget1


def test_lookup_by_name():
    '''Load two catalogs into the shelf and call an algorithm from the second'''
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1 = NamedWidget(name='widget1')
    catalog1.register(widget1)
    widget2 = NamedWidget(name='widget2')
    catalog2.register(widget2)
    dut = CatalogShelf(catalog1, catalog2)
    assert dut.lookup_by_name('widget1') == widget1
    assert dut.lookup_by_name('widget2') == widget2


def test_lookup_by_name_conflict():
    '''Load two catalogs into the shelf and call an algorithm that is in both'''
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1 = NamedWidget(name='conflicting_widget', tags={'version': ['one']})
    catalog1.register(widget1)
    widget2 = NamedWidget(name='conflicting_widget', tags={'version': ['two']})
    catalog2.register(widget2)
    dut = CatalogShelf(catalog1, catalog2)
    assert dut.lookup_by_name('conflicting_widget').tags['version'][0] == 'one'


def test_lookup_by_tags_conflicting_names() -> None:
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1 = NamedWidget(name='conflicting_widget', tags={'version': ['one']})
    catalog1.register(widget1)
    widget2 = NamedWidget(name='conflicting_widget', tags={'version': ['one'],
                                                           'label': ['another']})
    catalog2.register(widget2)
    dut = CatalogShelf(catalog1, catalog2)
    res = dut.lookup_by_tag_and(version='one')
    assert len(res) == 1
    got = res['conflicting_widget']
    assert isinstance(got, NamedWidget)
    assert 'label' not in got.tags


def test_tagvals() -> None:
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1 = NamedWidget(name='widget1', tags={'version': ['one']})
    catalog1.register(widget1)
    widget2 = NamedWidget(name='widget2', tags={'version': ['two'],
                                                'label': ['another']})
    catalog2.register(widget2)
    dut = CatalogShelf(catalog1, catalog2)
    res = dut.tagvals('version')
    assert res == {'one', 'two'}


def test_tagtypes() -> None:
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1 = NamedWidget(name='widget1', tags={'version': ['one']})
    catalog1.register(widget1)
    widget2 = NamedWidget(name='widget2', tags={'version': ['two'],
                                                'label': ['another']})
    catalog2.register(widget2)
    dut = CatalogShelf(catalog1, catalog2)
    res = dut.tagtypes
    assert res == {'label', 'version'}


def test_prepend() -> None:
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1 = NamedWidget(name='conflict', tags={'version': ['one']})
    catalog1.register(widget1)
    widget2 = NamedWidget(name='conflict', tags={'version': ['two']})
    catalog2.register(widget2)
    dut = CatalogShelf(catalog1, catalog2)

    got1 = dut.lookup_by_name('conflict')
    assert got1.tags == {'version': ['one']}

    dut.prepend(catalog2)
    got2 = dut.lookup_by_name('conflict')
    assert got2.tags == {'version': ['two']}


def test_register():
    '''Register a widget into a shelf with an empty catalog at the start.'''
    catalog1 = WidgetCatalog()
    catalog2 = WidgetCatalog()
    widget1_1 = NamedWidget(name='widget1', tags={'version': ['one']})
    catalog2.register(widget1_1)
    widget1_2 = NamedWidget(name='widget1', tags={'version': ['two']})
    dut = CatalogShelf(catalog1, catalog2)
    dut.register(widget1_2)
    assert dut.lookup_by_name('widget1').tags['version'][0] == 'two'


def test_register_fail():
    '''Running register() on a shelf should fail if the first catalog is not a MemoryCatalog.'''
    catalog1 = WidgetNonMemoryCatalog()
    catalog2 = WidgetCatalog()
    widget1_1 = NamedWidget(name='widget1', tags={'version': ['one']})
    catalog2.register(widget1_1)
    widget1_2 = NamedWidget(name='widget1', tags={'version': ['two']})
    dut = CatalogShelf(catalog1, catalog2)
    with pytest.raises(AssertionError, match='first catalog'):
        dut.register(widget1_2)
    looked_up = dut.lookup_by_name('widget1')
    assert isinstance(looked_up, NamedWidget)
    assert looked_up.tags['version'][0] == 'one'



'''Logging control object for the project.

Use this like this:
import logging
from ...wranger.logger import Logger

logger = Logger(__file__, level=logging.INFO, to_file=False, to_stdout=True)
logger.debug('message %s', arg)
'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
import sys
import logging
from typing import List, Optional, Tuple, Union


class Level():
    '''Alternative to log level constants, introduces VERBOSE'''
    CRITICAL: int = logging.CRITICAL
    ERROR: int = logging.ERROR
    WARN: int = logging.WARN
    INFO: int = logging.INFO
    DEBUG: int = logging.DEBUG
    VERBOSE: int = 5


class Logger:
    '''Logging control object for the project.

    Use like this:

    .. code-block:: python

        import logging
        from ...wrangler.logger import Logger

        logger = Logger(__file__, level=logging.INFO, to_file=False, to_stdout=True)
        logger.debug('message %s', arg)
    '''

    @classmethod
    def set_global_level(cls, level: int) -> None:
        '''Override log level for all loggers in this module.'''
        loggers = [logging.getLogger(name)
                   for name in logging.root.manager.loggerDict]  # pylint: disable=no-member,line-too-long
        loggers.append(logging.getLogger())  # Add root logger
        for logger in loggers:
            logger.setLevel(level)

    def __init__(self,
                 from_filename: str,
                 to_filename: str = 'univ',
                 level: int = logging.DEBUG,
                 to_stdout: bool = False,
                 to_file: bool = True,
                 only_if_filepath_contains: Optional[Union[str, List[str]]] = None,
                 only_in_line_range: Optional[Tuple[int, int]] = None):
        pfile = Path(from_filename)
        self._logger = logging.getLogger(pfile.name)
        self._logger.setLevel(level)

        if only_if_filepath_contains:
            self._logger.addFilter(FilepathContains(contains=only_if_filepath_contains))

        if only_in_line_range:
            self._logger.addFilter(OnlyInLineRange(line_range=only_in_line_range))

        formatter = logging.Formatter(
            '[%(asctime)s] %(levelname)s [%(filename)s.%(funcName)s:%(lineno)d] %(message)s',
            datefmt='%a, %d %b %Y %H:%M:%S')
        if to_file:
            fh = logging.FileHandler(f'{to_filename}.log')
            fh.setFormatter(formatter)
            self._logger.addHandler(fh)
        if to_stdout:
            sh = logging.StreamHandler(sys.stdout)
            sh.setFormatter(formatter)
            self._logger.addHandler(sh)

    def logger(self) -> logging.Logger:
        '''Access the logging.Logger associated with this object.'''
        return self._logger


class FilepathContains(logging.Filter):
    '''Logging filter that preserves logs whose filepath contains a given string.

    If multiple strings are provided, will preserve logs whose source file match
    any one of them.
    '''
    _contains: List[str]

    def __init__(self, *args, contains: Union[str, List[str]], **kwargs):
        if isinstance(contains, str):
            contains = [contains]
        self._contains = contains
        super().__init__(*args, **kwargs)

    def filter(self, record: logging.LogRecord) -> bool:
        return any(s in record.pathname for s in self._contains)


class OnlyInLineRange(logging.Filter):
    '''Logging filter that preserves logs who originate from a range of lines.'''
    _line_range: Tuple[int, int]

    def __init__(self, *args, line_range: Tuple[int, int], **kwargs):
        self._line_range = line_range
        super().__init__(*args, **kwargs)

    def filter(self, record: logging.LogRecord) -> bool:
        return self._line_range[0] <= record.lineno <= self._line_range[1]



'''Tests for wide_to_long.py.'''
import pandas as pd

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..wrangler.dataset import Column, Dataset, RoleName, Metadata
from .wide_to_long import WideToLongModel
# pylint: disable=missing-function-docstring,duplicate-code


def test_wide_to_long_sunny_day() -> None:
    model = WideToLongModel()
    dut = model.instantiate()

    olddf = pd.DataFrame({
        'timestamp': [1, 2, 3, 4],
        'col_a': ['a1', 'a2', 'a3', 'a4'],
        'col_b': ['b1', 'b2', 'b3', 'b4'],
    })
    metadata = Metadata(
        roles={
            RoleName.TIME: [
                Column('timestamp')],
            RoleName.TARGET: [
                Column('col_a'),
                Column('col_b')],
        })
    old = Dataset(metadata=metadata)
    old.dataframe = olddf

    got = dut.predict(old)
    gotdf = got.dataframe
    gotdf.sort_values(by=['timestamp', 'y'])

    wantdf = pd.DataFrame({
        'timestamp': [1, 2, 3, 4, 1, 2, 3, 4],
        'unique_id': ['col_a', 'col_a', 'col_a', 'col_a',
                      'col_b', 'col_b', 'col_b', 'col_b'],
        'y': ['a1', 'a2', 'a3', 'a4',
              'b1', 'b2', 'b3', 'b4'],
    })
    wantdf.sort_values(by=['timestamp', 'y'])

    pd.testing.assert_frame_equal(gotdf, wantdf)

    got_roles = got.metadata.roles
    assert got_roles[RoleName.TIME] == [
        Column('timestamp')]
    assert got_roles[RoleName.TIMESERIES_ID] == [
        Column('unique_id')]
    assert got_roles[RoleName.TARGET] == [
        Column('y')]


def test_already_long() -> None:
    dut = WideToLongModel().instantiate()

    old_df = pd.DataFrame({
        'timestamp': [1, 2, 3, 4, 1, 2, 3, 4],
        'variable': ['col_a', 'col_a', 'col_a', 'col_a',
                     'col_b', 'col_b', 'col_b', 'col_b'],
        'value': ['a1', 'a2', 'a3', 'a4',
                  'b1', 'b2', 'b3', 'b4'],
    })

    metadata = Metadata(roles={
        RoleName.TIME: [
            Column('timestamp')],
        RoleName.TIMESERIES_ID: [
            Column('variable')],
        RoleName.TARGET: [
            Column('value')],
    })

    old_dataset = Dataset(metadata=metadata)
    old_dataset.dataframe = old_df

    output = dut.predict(old_dataset)
    got_df = output.get_dataframe()
    pd.testing.assert_frame_equal(got_df, old_df)

    assert output.metadata == metadata



'''File defining global pytest fixtures.'''
import time
from typing import Iterable

import pytest

from .algorithms.impl.distributed_algorithm_instance import DistributedAlgorithmInstance
from .wrangler.logger import Level, Logger

log = Logger(__file__, to_stdout=False, level=Level.DEBUG).logger()


class Waiter():
    '''Wait until every distributed instance has been fit at least once, and none are locked.

    Also return if max_time has passed.

    Returns False if timed out, True if everything converged.

    Wait at least min_time.

    Convergence threshold is the min number of times we considered fitting
        on a unchanged state in order to be considered "converged".

    Only used by distributed algorithms.
    '''

    def _stop(self,
              models: Iterable[DistributedAlgorithmInstance],
              convergence_check: bool,
              pending_message_check: bool) -> bool:

        log.debug('CHECKING STOP CONDITON')

        models_copy = list(models)

        if not models_copy:
            log.debug('NO MODELS TO CHECK')

        locked = [m.locked for m in models_copy]
        trained = [m.trained for m in models_copy]
        converged = [m.converged for m in models_copy]
        pending_messages = [m._pending_message_from_neighbor for m in models_copy]  # pylint: disable=protected-access

        log.debug('locked models: %s', locked)
        log.debug('trained models: %s', trained)
        log.debug('pending message: %s', pending_messages)
        log.debug('converged: %s', converged)

        retval = (not any(locked)) and all(trained)
        if convergence_check:
            # We are specifically dropping locked as a check for convergence.
            # This allows for deciders that keep sending even after everybody has converged.
            retval = all(trained) and all(converged)

        if pending_message_check:
            retval = retval and not any(pending_messages)

        if retval:
            log.debug('STOP CONDITION MET')

        return retval

    def __call__(self,
                 models: Iterable[DistributedAlgorithmInstance],
                 interval: float = 0.5,
                 max_time: float = 20.0,
                 min_time: float = 0.5,
                 convergence_check: bool = False,
                 pending_message_check: bool = False) -> bool:
        log.log(Level.VERBOSE, 'STARTING WAITER')

        start = time.monotonic()

        time.sleep(min_time)

        log.log(Level.VERBOSE, 'ABOUT TO START LOOP')

        models_copy = list(models)

        while not self._stop(models_copy, convergence_check, pending_message_check):

            elapsed = time.monotonic() - start
            if elapsed > max_time:
                log.warning('Waiter timed out.')
                return False
            time.sleep(interval)

        return True


@pytest.fixture(scope='session')
def wait_til_all_fit() -> Waiter:
    '''Serves up a waiting function.

    Only used by distributed algorithms.
    '''
    return Waiter()


# TODO(Merritt): Clearer takes *args instead of a list
class Clearer():
    '''Asserts that there are no exceptions left in the exception queue for any model.

    If there are any left, provides a helpful error message.

    Only used by distributed algorithms.
    '''

    def __call__(
        self,
        models: Iterable[DistributedAlgorithmInstance]
    ) -> None:

        found = ''
        for model in models:
            next_exc = model.poll_exceptions()
            while next_exc is not None:
                found += f'{next_exc}'
                next_exc = model.poll_exceptions()
        assert found == ''  # There should be no additional exceptions


@pytest.fixture(scope='session')
def assert_no_exceptions() -> Clearer:
    '''Asserts that there are no elements left in the exception queue.

    If there are any left, provides a helpful error message.

    Only used by distributed algorithms.
    '''
    return Clearer()



'''Tests for sklearn_model.py.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Tuple

import numpy as np
import pandas as pd
import pytest
from sklearn import datasets  # type: ignore[import-untyped]


from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator
from ...metrics.impl.metric_auto import MetricCatalogAuto
from ...problem_def.hyperparam_config import HyperparamConfig
from ...problem_def.task import TaskType
from ...searcher.searcher import SearcherImpl
from ...tables.impl.table import Table
from ...tables.impl.table_auto import TableCatalogAuto
from ...wrangler.dataset import Column, Dataset, Metadata, RoleName
from ..connect import ConnectorModel
from ..impl.algorithm import Algorithm
from ..impl.algorithm_auto import AlgorithmCatalogAuto
from .sklearn_algorithm import SklearnAlgorithmInstance

# pylint: disable=missing-function-docstring, disable=super-init-not-called,duplicate-code
_ = TableCatalogAuto()  # pylint: disable=pointless-statement


def load_regression_dataset() -> Tuple[Dataset, Dataset, Dataset]:
    # Load the diabetes dataset
    diabetes_x, diabetes_y = datasets.load_diabetes(return_X_y=True)

    # Use only one feature
    diabetes_x = diabetes_x[:, np.newaxis, 2]

    # Split the data into training/testing sets
    diabetes_x_train = diabetes_x[:-20]
    diabetes_x_test = diabetes_x[-20:]

    # Split the targets into training/testing sets
    diabetes_y_train = diabetes_y[:-20]
    diabetes_y_test = diabetes_y[-20:]

    dataset_train = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        ),
        covariates=pd.DataFrame(diabetes_x_train, columns=['attribute']),
        target=pd.DataFrame({'target': diabetes_y_train})
    )
    dataset_test = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        ),
        covariates=pd.DataFrame(diabetes_x_test, columns=['attribute'])
    )
    ground_truth = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        ),
        ground_truth=pd.DataFrame({'target': diabetes_y_test})
    )
    return (dataset_train, dataset_test, ground_truth)


def load_classification_dataset() -> Tuple[Dataset, Dataset, Dataset]:
    # Load the breast cander dataset
    bc_x_full, bc_y_series = datasets.load_breast_cancer(return_X_y=True, as_frame=True)
    assert isinstance(bc_x_full, pd.DataFrame)
    assert isinstance(bc_y_series, pd.Series)
    bc_y = pd.DataFrame({'target': bc_y_series})

    # restrict number of attributes for wider variability of results
    bc_x = bc_x_full.iloc[:, :3]

    test_size = 50
    # Split the data into training/testing sets
    bc_x_train = bc_x[:-test_size]
    bc_x_test = bc_x[-test_size:]

    # Split the targets into training/testing sets
    bc_y_train = bc_y[:-test_size]
    bc_y_test = bc_y[-test_size:]

    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.BINARY_CLASSIFICATION
    )

    dataset_train = Dataset(
        metadata=metadata,
        covariates=bc_x_train,
        target=bc_y_train
    )

    dataset_test = Dataset(
        metadata=metadata,
        covariates=bc_x_test
    )
    ground_truth = Dataset(
        metadata=metadata,
        ground_truth=bc_y_test
    )
    return (dataset_train, dataset_test, ground_truth)


@pytest.mark.parametrize("name,value,precision", [
    ('sklearn.linear_model.LinearRegression', 2548.0723, 1e-6),
    ('sklearn.svm.SVR', 3252.0655, 1e-6),
    ('sklearn.svm.LinearSVR', 5089.4231, 1e-6),
    ('sklearn.ensemble.GradientBoostingRegressor', 2928.0825, 1e-6),
    ('sklearn.ensemble.RandomForestRegressor', 4641.0977, 1e-6),
    ('sklearn.linear_model.Ridge', 3604.2696, 1e-6),
    ('sklearn.linear_model.Lasso', 3482.7721, 1e-6),
    ('sklearn.linear_model.ElasticNet', 5551.8319, 1e-6),
    ('sklearn.linear_model.LassoCV', 2549.1403, 1e-6),
    ('sklearn.ensemble.AdaBoostRegressor', 2884.9437, 1e-6),
    ('sklearn.tree.ExtraTreeRegressor', 6374.6676, 1e-6),
    ('sklearn.ensemble.BaggingRegressor', 4861.7033, 1e-6),
])
def test_regression_sunny_day(name: str, value: float, precision: float) -> None:
    catalog = AlgorithmCatalogAuto()
    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    model = catalog.lookup_by_name(name)
    instance = model.instantiate()

    train, test, ground_truth = load_regression_dataset()

    instance.fit(train)

    result = instance.predict(test)
    assert metric.calculate(pred=result,
                            ground_truth=ground_truth
                            ) == pytest.approx(value, precision)


@pytest.mark.parametrize('name,value,precision', [
    ('sklearn.linear_model.LogisticRegression', 0.88, 1e-6),
    ('sklearn.ensemble.RandomForestClassifier', 0.94, 1e-6),
    ('sklearn.svm.SVC', 0.9, 1e-6),
    ('sklearn.svm.LinearSVC', 0.94, 1e-6),
    ('sklearn.naive_bayes.MultinomialNB', 0.86, 1e-6),
    ('sklearn.discriminant_analysis.LinearDiscriminantAnalysis', 0.92, 1e-6),
    ('sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis', 0.92, 1e-6),
    ('sklearn.ensemble.AdaBoostClassifier', 0.88, 0.03),
    ('sklearn.tree.ExtraTreeClassifier', 0.92, 1e-6),
    ('sklearn.ensemble.BaggingClassifier', 0.9, 1e-6),
    ('sklearn.linear_model.PassiveAggressiveClassifier', 0.48, 1e-6),
    ('sklearn.neural_network.MLPClassifier', 0.94, 1e-6),
    ('sklearn.ensemble.GradientBoostingClassifier', 0.92, 1e-6),
])
def test_classification_sunny_day(name: str, value: float, precision: float) -> None:
    catalog = AlgorithmCatalogAuto()
    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    model: Algorithm = catalog.lookup_by_name(name)
    instance = model.instantiate()

    train, test, ground_truth = load_classification_dataset()

    instance.fit(train)

    result = instance.predict(test)
    assert metric.calculate(pred=result,
                            ground_truth=ground_truth
                            ) == pytest.approx(value, precision)


class FakeSklearnImplementation:
    '''A fake Sklearn model implementation for testing purpose.'''

    def fit(self, _, __) -> None:
        return None

    def predict(self, _) -> np.ndarray:
        return np.array(["predict"])

    def predict_proba(self, _) -> np.ndarray:
        return np.array([["predict_proba", "predict_proba"],
                         ["predict_proba_row2", "predict_proba_row2"],
                         ["predict_proba_row3", "predict_proba_row3"]])


class FakeSklearnModelInstance(SklearnAlgorithmInstance):
    ''' Fake Sklearn model for testing purpose. '''

    def __init__(self):
        self._impl = FakeSklearnImplementation()
        self._algorithm = ConnectorModel()


def test_probability_output():
    fake_metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.BINARY_CLASSIFICATION)
    fake_dataset = Dataset(metadata=fake_metadata)
    fake_dataset['covariates'] = ['covariates']
    fake_dataset['target'] = ['target']
    model = FakeSklearnModelInstance()
    model.fit(fake_dataset)

    got = model.predict(fake_dataset)

    assert 'predictions' in got
    assert isinstance(got.predictions, pd.DataFrame)
    assert got.predictions.shape == (1, 1)
    assert got.predictions.iat[0, 0] == 'predict'

    assert 'probabilities' in got
    assert isinstance(got.probabilities, Table)
    assert got.probabilities.shape == (3, 1)
    assert got.probabilities.as_(pd.DataFrame).iat[0, 0] == 'predict_proba'  # type: ignore[attr-defined] # pylint: disable=line-too-long


def test_probability_output_for_non_binary_classification_task():
    ''' We only want to get probabilities output if it's a binary
        classification task.
    '''
    fake_metadata_regression = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.REGRESSION)
    fake_dataset = Dataset(metadata=fake_metadata_regression)
    fake_dataset['covariates'] = ['covariates']
    fake_dataset['target'] = ['target']
    model = FakeSklearnModelInstance()
    model.fit(fake_dataset)

    got = model.predict(fake_dataset)

    assert 'probabilities' not in got

    assert 'predictions' in got
    assert isinstance(got.predictions, pd.DataFrame)
    assert got.predictions.shape == (1, 1)
    assert got.predictions.iat[0, 0] == 'predict'


@pytest.mark.parametrize('name,want_number,want_des', [
    ('sklearn.linear_model.LogisticRegression',
     8,
     'test_pipe@sklearn.linear_model.logisticregression:c=0.1,class_weight=balanced'),
    ('sklearn.ensemble.RandomForestClassifier',
     24,
     'test_pipe@sklearn.ensemble.randomforestclassifier:'
     'class_weight=balanced,max_depth=none,min_samples_split=10'),
    ('sklearn.svm.SVC',
     10,
     'test_pipe@sklearn.svm.svc:c=0.01,class_weight=none'),
    ('sklearn.svm.LinearSVC',
     10,
     'test_pipe@sklearn.svm.linearsvc:c=0.01,class_weight=none'),
    ('sklearn.tree.ExtraTreeClassifier',
     24,
     'test_pipe@sklearn.tree.extratreeclassifier:'
     'class_weight=balanced,max_depth=8,min_samples_split=2'),
    ('sklearn.ensemble.GradientBoostingClassifier',
     90,
     'test_pipe@sklearn.ensemble.gradientboostingclassifier:'
     'max_depth=8,max_features=none,min_samples_leaf=2,min_samples_split=10'),
    ('sklearn.ensemble.GradientBoostingRegressor',
     90,
     'test_pipe@sklearn.ensemble.gradientboostingregressor:'
     'max_depth=10,max_features=none,min_samples_leaf=2,min_samples_split=10'),
    ('sklearn.svm.SVR',
     5,
     'test_pipe@sklearn.svm.svr:c=0.01'),
    ('sklearn.svm.LinearSVR',
     5,
     'test_pipe@sklearn.svm.linearsvr:c=0.1'),
    ('sklearn.ensemble.RandomForestRegressor',
     12,
     'test_pipe@sklearn.ensemble.randomforestregressor:max_depth=8,min_samples_split=2'),
    ('sklearn.linear_model.Ridge',
     5,
     'test_pipe@sklearn.linear_model.ridge:alpha=0.001'),
    ('sklearn.linear_model.Lasso',
     5,
     'test_pipe@sklearn.linear_model.lasso:alpha=5.0'),
    ('sklearn.tree.ExtraTreeRegressor',
     12,
     'test_pipe@sklearn.tree.extratreeregressor:max_depth=10,min_samples_split=2')
])
def test_grid_search(name: str, want_number: int, want_des: str) -> None:
    # make boundpipeline w random forest clf
    catalog = AlgorithmCatalogAuto()
    algorithm = catalog.lookup_by_name(name)
    bound_pipeline = BoundPipeline(name='test_pipe')
    bound_pipeline.step(model=algorithm)
    searcher = SearcherImpl(HyperparamConfig(clause={'hyperparams': []}))
    got = searcher.bind_hyperparams(bound_pipeline)
    assert len(got) == want_number
    assert Designator(want_des) in got



'''Contains ProblemDefinition'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import json
import textwrap
from typing import Any, Dict, List, Optional, Set, Union

from aenum import Enum  # type: ignore[import-untyped]

from ..config_components.impl.config_component import (
    ConfigComponent, ConfigError, InvalidKeyError,
    MissingKeyError, ParsingErrors, ValidationErrors)
from ..config_components.impl.config_component_catalog import ConfigComponentCatalog
from ..config_components.impl.config_component_auto import ConfigComponentAuto
from ..wrangler.constants import ProblemDefKeys
from .aggregation_config import AggregationConfig, AggregationError
from .cross_validation_config import CrossValidationConfig
from .hyperparam_config import HyperparamConfig, HyperparamError
from .metric_config import MetricConfig, MetricConfigError
from .output_config import OutputConfig
from .problem_def_interface import ProblemDefInterface
from .task import Task

# pylint: disable=too-many-branches,too-many-statements


class ProblemDefinition(ProblemDefInterface):
    '''Represents information needed to define an AutoML run'''
    _catalog: Optional[ConfigComponentCatalog] = None
    _plugin_components: Dict[str, ConfigComponent]
    _task: Optional[Task] = None
    _aggregation_config: Optional[AggregationConfig] = None
    _cross_validation_config: Optional[CrossValidationConfig] = None
    _metric_configs: Optional[Dict[str, MetricConfig]] = None
    _hyperparam_config: Optional[HyperparamConfig] = None
    _output_config: Optional[OutputConfig] = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_catalog']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._catalog = ConfigComponentAuto()

    def allowed_keys(self) -> Set[str]:
        '''Set of keys that are acceptable for this clause'''
        allowed = super().allowed_keys()
        allowed.update(self._plugin_components.keys())
        return allowed

    def required_keys(self) -> Set[str]:
        '''Subset of ALLOWED_KEYS that are required'''
        return {self.Keys.TASK.value, self.Keys.DATASET.value}   # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long

    def __init__(self,
                 clause: Union[str, Dict[str, Any]]):

        if self._catalog is None:
            self._catalog = ConfigComponentAuto()

        if isinstance(clause, str):
            clause = self._parse_json(pdef=clause)
        super().__init__(clause=remove_comments(clause))

        errors: List[ConfigError] = []

        self._plugin_components = {}

        for plugin_name, constructor in self._catalog.items():
            try:
                plugin_clause = self._get(plugin_name)
            except MissingKeyError:
                # We do not require all plugin clauses to be present.
                plugin_clause = {}

            try:
                # But we still want to instantiate the plugin so that we can use its validator.
                self._plugin_components[plugin_name] = (
                    constructor(clause=plugin_clause, problem_def=self))
            except ParsingErrors as err:
                errors.extend(err.errors)

        try:
            self._task = Task(self._get(
                ProblemDefKeys.TASK))
        except KeyError as err:
            errors.append(InvalidKeyError(err))

        try:
            self._cross_validation_config = CrossValidationConfig(self._get_with_default(
                ProblemDefKeys.CROSS_VALIDATION, dflt={}))
        except KeyError as err:
            errors.append(InvalidKeyError(err))

        try:
            metrics_clause = self._get_with_default(ProblemDefKeys.METRICS, dflt={})
            self._metric_configs = {name: MetricConfig(clause={name: info})
                                    for name, info in metrics_clause.items()}
        except MetricConfigError as err:
            errors.append(err)

        try:
            hyperparam_clause = self._get_with_default(ProblemDefKeys.HYPERPARAMS, dflt=[])
            self._hyperparam_config = HyperparamConfig(
                clause={ProblemDefKeys.HYPERPARAMS.value: hyperparam_clause})
        except HyperparamError as err:
            errors.extend(err.errors)

        try:
            output_clause = self._get_with_default(ProblemDefKeys.OUTPUT, dflt={})
            self._output_config = OutputConfig(output_clause)
            # TODO(piggy): Confirm that OutputConfig calls its own validate()
            # and drop this call.
            self._output_config.validate()
        except ValidationErrors as err:
            errors.extend(err.errors)

        try:
            aggregation_clause = self._get_with_default(ProblemDefKeys.AGGREGATION, dflt={})
            self._aggregation_config = AggregationConfig(aggregation_clause)
        except AggregationError as err:
            errors.extend(err.errors)

        try:
            # This is where we do validation between sections.
            self.validate()
        except ValidationErrors as err:
            errors.extend(err.errors)

        if len(errors) > 0:
            raise ValidationErrors(errors)

    def _parse_json(self, pdef: str) -> Dict[str, Any]:
        '''Attempts to interpret pdef as a JSON string or path to JSON file.

        If neither succeeds, raises ParsingErrors.
        '''
        errors: List[ConfigError] = []

        try:
            retval = json.loads(pdef)
            return retval
        except json.decoder.JSONDecodeError as err:
            errors.append(ConfigError(err))

        try:
            with open(pdef, 'r', encoding='utf8') as f:
                retval = json.load(f)
            return retval
        except FileNotFoundError as err:
            errors.append(ConfigError(err))
        except OSError as err:
            errors.append(ConfigError(err))
        except json.decoder.JSONDecodeError as err:
            errors.append(ConfigError(err))

        errors.insert(0, ConfigError(
            'String provided could not be parsed as JSON or path to JSON file.'))
        raise ParsingErrors(errors=errors)

    def __str__(self):
        return textwrap.dedent(f'''
            problem_type={self.task}
        ''')

    def validate(self, **kwargs) -> None:
        super().validate(**kwargs)  # all errors thrown by base class are fatal

        errors: List[ConfigError] = []

        try:
            self.task.validate()
        except ValidationErrors as err:
            # If task type is not provided, we cannot continue with validation
            raise err

        if self._metric_configs is not None:
            for metric_config in self._metric_configs.values():
                try:
                    metric_config.validate()
                except MetricConfigError as err:
                    errors.append(err)

        for component in self._plugin_components.values():
            try:
                component.validate(problem_def=self)
            except ValidationErrors as errs:
                errors.extend(errs.errors)

        if len(errors) > 0:
            raise ValidationErrors(errors)

    @property
    def task(self) -> Task:
        '''Explains what kind of problem we're solving.

        This affects the selection of pipeline templates and models.
        '''
        assert self._task is not None, (
            'BUG: attempt to access problem_type but it is None.')
        return self._task

    @property
    def metric_configs(self) -> Dict[str, MetricConfig]:
        '''Names and other information, if applicable, about all the metrics for this problem.

        If there are no metrics specified, will return an empty list, in which case the wrangler is
        responsible for using a default metric based on the problem type.
        '''
        assert self._metric_configs is not None, (
            'BUG: attempt to access metric_config but it is None.')
        return self._metric_configs

    @property
    def cross_validation_config(self) -> CrossValidationConfig:
        '''
        Get configuration information for cross-validation.
        '''
        assert self._cross_validation_config is not None, (
            'BUG: attempt to access cross_validation_config but it is None.'
        )

        return self._cross_validation_config

    @property
    def output(self) -> OutputConfig:
        '''All the configuration about things to output.'''
        assert self._output_config is not None, (
            'BUG: empty output session should be handled in constructor.'
        )
        return self._output_config

    @property
    def hyperparams(self) -> HyperparamConfig:
        '''Configuration for all hyperparam overrides.'''
        assert self._hyperparam_config is not None, (
            'BUG: self._hyperparam_config is supposed to be set in __init__().')
        return self._hyperparam_config

    @property
    def aggregation(self) -> AggregationConfig:
        '''Configuration for rank aggregation.

        The only valid entry is "method" which gives
        the catalog name of a rank aggregation method.
        See the top level "aggregators" directory.
        '''
        assert self._aggregation_config is not None, (
            'BUG: self._aggregation_config is supposed to be set in __init__().')
        return self._aggregation_config

    def get_conf(self, config_name: Union[str, Enum]) -> ConfigComponent:
        '''Get the plugin ConfigComponent for the config_name.'''
        if isinstance(config_name, Enum):
            config_name = config_name.value
        assert isinstance(config_name, str)
        return self._plugin_components[config_name]


def remove_comments(entry: Any) -> Any:
    '''Remove '_comments' fields from every dict inside an anonymous object.'''
    if isinstance(entry, dict):
        retval = {
            k: remove_comments(v)
            for k, v in entry.items()
            if k != '_comments'
        }
        return retval
    if isinstance(entry, list):
        return [remove_comments(v) for v in entry]
    return entry



'''Discoverer that uses static configuration.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import logging
from queue import Queue
from typing import Dict, List

import numpy as np

from ..communicators.impl.communicator import Communicator
from ..config_components.distributed_config import DistributedConfig
from ..neighbor_manager.event import NewNeighbor
from ..neighbor_manager.node_id import NodeID
from .impl.discoverer import Discoverer
from .impl.discoverer_catalog import DiscovererCatalog

from ..wrangler.logger import Logger

log = Logger(__file__, level=logging.DEBUG).logger()


class StaticDiscoverer(Discoverer):
    '''Discoverer that uses static configuration.'''
    name = 'static'
    tags: Dict[str, List[str]] = {}

    def __init__(self, config: DistributedConfig,
                 communicator: Communicator):
        super().__init__(config=config, communicator=communicator)
        neighbors = self._config.get_static_adjacency(my_id=communicator.my_id)
        assert neighbors is not None, (
            'BUG: Attempt to define StaticDiscoverer with no static config.')
        self._neighbors = neighbors
        known = self._communicator.known_neighbors
        log.debug('StaticDiscoverer: known_neighbors=%s, adding=%s', known, neighbors)
        known.update(neighbors)

    def start(self, queue: Queue):
        for node in self._neighbors:
            queue.put(NewNeighbor(neighbor=NodeID(node)))
        self._communicator.start(
            queue=queue
        )

    def laplacian(self) -> List[List[int]]:
        '''Convert the row adjacency config to a Laplacian matrix.

        # https://en.wikipedia.org/wiki/Laplacian_matrix#:~:text=Definitions%20for%20simple%20graphs,-%5Bedit%5D
        '''
        row_adjacency = self._config.get_adjacency()
        assert row_adjacency is not None, (
            'BUG: Attempt to get Laplacian with no adjacency config.')

        max_id = max(k for k in row_adjacency.keys())
        degree = np.zeros((max_id, max_id), dtype=int)
        adjacency = np.zeros((max_id, max_id), dtype=int)
        for i in range(max_id):
            i_node = NodeID(i + 1)
            if i_node not in row_adjacency.keys():
                continue
            degree[i][i] = len(row_adjacency[i_node])
            for j in row_adjacency[i_node]:
                adjacency[i][j - 1] = 1

        return (degree - adjacency).tolist()


def register(catalog: DiscovererCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(StaticDiscoverer)



'''Tests for simple_imputer.'''
from typing import Optional
import pytest

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import numpy as np
import pandas as pd

from ....wrangler.dataset import Dataset

from .simple_imputer import SimpleImputerModel
# pylint: disable=missing-function-docstring, protected-access,duplicate-code


def test_sunny_day() -> None:
    instance = SimpleImputerModel().instantiate()

    # Simple example dataset from:
    # https://vitalflux.com/imputing-missing-data-sklearn-simpleimputer/#SimpleImputer_Python_Code_Example

    students = [[85, 'M', 'verygood'],
                [95.0, 'F', 'excellent'],
                [75, None, 'good'],
                [np.nan, 'M', 'average'],
                [70, 'M', 'good'],
                [np.nan, None, 'verygood'],
                [92, 'F', 'verygood'],
                [98, 'M', 'excellent']]
    students_df = pd.DataFrame(students, columns=['marks', 'gender', 'result'])

    students_dataset = Dataset()
    students_dataset.dataframe = students_df

    result: Optional[Dataset] = instance.predict(students_dataset)
    assert result is not None
    result_df: pd.DataFrame = result.dataframe

    # we want default strategy to be 'mean' for numerical values (in this case, ~85.83)
    # and 'most_frequent' for categorical values (in this case, 'M')
    assert set(result_df.columns) == set(['marks', 'gender', 'result'])
    assert pytest.approx(85.8333333, 1e-6) == result_df.iat[3, 0]
    assert 'M' == result_df.iat[2, 1]



'''Interface to metadata for a neighbor node.

This augments the NeighborState object.
'''
import abc
from typing import Any, Dict, Optional

from .neighbor_state_interface import NeighborStateInterface


class ForceFalse(Exception):
    '''Force the decider to return False.'''


class NeighborMetadataInterface(metaclass=abc.ABCMeta):
    '''Interface for metadata for a neighbor node.'''

    _current_state: Optional[NeighborStateInterface] = None
    _scratch: Dict[str, Any]

    def __init__(self, my_state: Optional[NeighborStateInterface] = None):
        '''Initialize the metadata.'''
        super().__init__()
        self._current_state = my_state
        self._scratch = {}

    @property
    def scratch(self) -> Dict[str, Any]:
        '''Get the scratch space.

        This is a dictionary that can be used by deciders to store
        data that persists from one call to the next.

        The recommendation is to add a subdictionary keyed by the
        name of the decider, which can contain whatever the decider
        needs. Values from other deciders should be considered
        read only.
        '''
        return self._scratch

    @property
    @abc.abstractmethod
    def current_state(self) -> Optional[NeighborStateInterface]:
        '''The current state of the neighbor.'''

    @current_state.setter
    @abc.abstractmethod
    def current_state(self, state: Optional[NeighborStateInterface]):
        '''Set the current state of the neighbor.'''

    @property
    @abc.abstractmethod
    def last_state(self) -> Optional[NeighborStateInterface]:
        '''The last state of the neighbor.'''

    @property
    @abc.abstractmethod
    def last_state_update_time(self) -> float:
        '''The time the neighbor was last updated.'''

    @property
    @abc.abstractmethod
    def time_since_last_state_update(self) -> float:
        '''The time since the neighbor was last updated.'''

    @property
    @abc.abstractmethod
    def last_state_sent(self) -> Optional[NeighborStateInterface]:
        '''The last state that was sent to the neighbor.'''

    @last_state_sent.setter
    @abc.abstractmethod
    def last_state_sent(self, state: Optional[NeighborStateInterface]):
        '''Set the last state that was sent to the neighbor.'''

    @abc.abstractmethod
    def sent_current_state(self):
        '''Mark the current state as sent.'''

    @property
    @abc.abstractmethod
    def last_state_sent_time(self) -> float:
        '''The time the last state was sent to the neighbor.'''

    @property
    @abc.abstractmethod
    def time_since_last_state_sent(self) -> float:
        '''The time since the last state was sent to the neighbor.'''

    @property
    @abc.abstractmethod
    def cumulative_payload_sent(self) -> int:
        '''The total payload sent to the neighbor.'''



'''Distributed Mean'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# Linear and Logistic Regression have portions of identical code that should
# not be abstracted away.
# pylint: disable=duplicate-code,no-member

import pickle
from typing import Iterable, Optional, Tuple

import numpy as np
import pandas as pd

from ...algorithms.impl.distributed_algorithm_instance import NoDataError
from ...catalog.catalog import upcast
from ...problem_def.task import DataType
from ...wrangler.dataset import Dataset, DatasetKeys
from ...wrangler.logger import Logger

from ..impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..impl.linear_distributed_algorithm_instance import (
    LinearDistributedAlgorithmInstance, LinearNeighborState)


logger = Logger(__file__).logger()


class AutonMeanNeighbor(LinearNeighborState['AutonMeanNeighbor']):
    '''A neighbor in the distributed mean model.

    The elements of v are the mean of each column in the data.
    '''

    def encode(self) -> bytes:
        '''Encode message for distributed neighbors.'''
        return pickle.dumps((self._v, self._columns))

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'AutonMeanNeighbor':
        '''Decode a message from a neighbor.'''
        v, columns = pickle.loads(serialized_model)
        assert isinstance(v, np.ndarray), (
            f'BUG: expected v to be an np.ndarray, instead found {v} of type {type(v)}'
        )
        columns = cls._cast_columns(columns)
        return cls(v=v, columns=columns)


class AutonMeanInstance(LinearDistributedAlgorithmInstance):
    '''A distributed model that calculates the mean of each column in the data

    As a distributed model, this instance can share information about its
    trained state with other instances and update its trained state using
    information shared by other instances.
    '''
    _neighbor_constructor = AutonMeanNeighbor

    # This is a type annotation for _my_state.
    _predict_state: Optional[AutonMeanNeighbor] = None

    @property  # type: ignore[override]
    def _my_state(self) -> Optional[AutonMeanNeighbor]:
        retval = LinearDistributedAlgorithmInstance._my_state.fget(self)  # type: ignore[attr-defined] # pylint: disable=assignment-from-no-return, line-too-long
        assert retval is None or isinstance(retval, AutonMeanNeighbor), (
            'BUG: expected _my_state to be None or an AutonMeanNeighbor.')
        return retval

    @_my_state.setter
    def _my_state(self, value: Optional[AutonMeanNeighbor]) -> None:
        assert value is None or isinstance(value, AutonMeanNeighbor), (
            'BUG: expected value to be None or an AutonMeanNeighbor.')
        LinearDistributedAlgorithmInstance._my_state.fset(self, value)  # type: ignore[attr-defined]

    @property
    def _yhat_is_proba(self) -> bool:
        return False

    @property
    def _yhat_is_multivalued(self) -> bool:
        """True if yhat returns a vector of results, 1 per column of x."""
        return True

    @property
    def _neighbor_models_iter(self) -> Iterable[AutonMeanNeighbor]:
        for v in super()._neighbor_models_iter:
            assert isinstance(v, AutonMeanNeighbor), (
                'BUG: expected neighbor_models to contain AutonMeanNeighbor, '
                f'instead found {v} of type {type(v)}'
            )
            yield v

    def _yhat(self, x: np.ndarray) -> np.ndarray:
        assert self._predict_state is not None, (
            'BUG: self._predict_state should not be None when _yhat is called.'
            ' This should have been set in _fit'
        )
        return self._predict_state.v

    def _decode(self, serialized_model: bytes) -> AutonMeanNeighbor:
        '''Decode a message from distributed neighbors.'''
        return AutonMeanNeighbor.decode(serialized_model)

    def _prepare_data(self, dataset: Optional[Dataset]) -> Tuple[np.ndarray, np.ndarray]:
        '''Extract covariates and target, accounting for the possibility that there's no data.'''
        if dataset is None:
            if list(self._neighbor_models_iter):
                d = max(len(m.v) for m in self._neighbor_models_iter)
                np_cov = np.zeros((0, d))
            else:
                # TODO(Piggy/Dan) Need to figure out how to deal with no neighbors and no data.
                raise NoDataError("No data and no neighbors.")
            np_tar = np.zeros(0)
        else:
            dataset = dataset.sorted_columns()
            if DatasetKeys.DATAFRAME.value in dataset:
                orig_cov = dataset[DatasetKeys.DATAFRAME.value]
            elif DatasetKeys.COVARIATES.value in dataset:
                orig_cov = dataset[DatasetKeys.COVARIATES.value]
            else:
                # This is for the case of predict with an empty dataset.
                orig_cov = pd.DataFrame()

            np_tar = np.array([0])  # type: ignore[assignment]
            np_cov = np.array(orig_cov, dtype=float)  # type: ignore[assignment]

        return np_cov, np_tar

    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Fit a model based on train data.

        This sets self.trained to True.
        '''
        if dataset is None:
            if list(self._neighbor_models_iter):
                if self._columns is None:
                    # TODO(piggy/Dan): Consider the case of neighbors with different columns.
                    self._columns = next(iter(self._neighbor_models_iter)).columns
            else:
                return  # We have neither data nor neighbors: there is nothing to do.

        try:
            np_cov, _ = self._prepare_data(dataset)
        except NoDataError:
            return  # There is nothing to do.

        logger.debug('Fitting %s; dataset: %r, len(neighbor_models): %s',
                     type(self).__name__, dataset, len(list(self._neighbor_models_iter)))
        x = np_cov
        n, _ = x.shape

        # number of non-nan observations per column
        per_col_n = np.count_nonzero(~np.isnan(x), axis=0)

        omega = self._omega
        _lambda = self._lambda

        if self._my_state is None:
            # v is the mean of every column.
            self._my_state = AutonMeanNeighbor(v=np.nanmean(x, axis=0), columns=self._columns)
        elif self._my_state.columns is None:
            self._my_state.columns = self._columns

        assert isinstance(self._my_state, AutonMeanNeighbor), (
            'BUG: expected self._my_state to be a AutonMeanNeighbor, '
            f'instead found {self._my_state} of type {type(self._my_state)}'
        )
        k = len(self._neighbor_metadata)

        neighbor_vs = [np.array(m.v, dtype=float) for m in self._neighbor_models_iter]

        if n == 0:
            # If you have no data take the mean of your neighbors.
            # If there are also no neighbors do nothing.
            if k != 0:
                self._my_state.v = np.nanmean(neighbor_vs, axis=0)
        else:
            x_sum = np.nansum(x, axis=0)

            # nansum outputs 0.0 if a col is all nans, but we want nan.
            x_sum[per_col_n == 0] = np.nan

            # get v_old (v from last fit) if it exists or set it to 0
            v_old = self._my_state.v.copy()

            prev_self_coef = (1 - omega) / omega

            old_nan_indicator = np.array(~np.isnan(v_old), dtype=int)

            neighbor_non_nan_counts = (~np.isnan(np.array(neighbor_vs))).sum(axis=0)

            vs_to_sum = np.array(
                [x_sum, prev_self_coef * v_old] + [_lambda * nv for nv in neighbor_vs])

            # 1darray - for each column of vs_to_sum, True if all values are nan
            all_nan_cols = np.count_nonzero(~np.isnan(vs_to_sum), axis=0) == 0

            self._my_state.v = np.nansum(vs_to_sum, axis=0) / (
                per_col_n + (prev_self_coef * old_nan_indicator)
                + (_lambda * neighbor_non_nan_counts))

            # nansum outputs 0.0 if a col is all nans, but we want nan.
            self._my_state.v[all_nan_cols] = np.nan
        logger.debug('%s fit complete; v: %r', type(self).__name__, self._my_state.v)


class AutonMean(Algorithm):
    '''Class for Auton Lab's implementation of distributed mean.'''
    _name = "auton_mean"
    _tags = {
        'preprocessor': ['true'],
        'data_types': [DataType.TABULAR.name],
        'source': ['auton_lab'],
        'distributed': ['true']
    }
    _instance_constructor = AutonMeanInstance
    _default_hyperparams = {
        'L2': 0.0,  # Not used by mean, but needed for linear algorithms
        'Lambda': 1.0,
        'omega': 2.0 / 3.0,
        'tol': 0.000000001,
        'maxiter': None,
    }

    def instantiate(self, **hyperparams) -> 'AutonMeanInstance':
        return super().instantiate(**self.hyperparams(**hyperparams))


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = AutonMean(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Convert two-valued fields into binary fields.'''
import logging
from typing import Any, Dict, Optional


# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..catalog.catalog import upcast
from ..wrangler.dataset import Dataset, RoleName
from ..wrangler.logger import Logger

logger = Logger(__file__, level=logging.INFO, to_file=False, to_stdout=False).logger()


class BinarizerModelInstance(AlgorithmInstance):
    '''Connect one model to another by transforming names.'''
    _pos_labels: Dict[RoleName, Any]

    def __init__(self, parent, **hyperparams: Any):
        super().__init__(parent=parent)

        self._pos_labels = {
            RoleName[n.upper()]: v
            for n, v in self.algorithm.hyperparams(**hyperparams).items()
        }

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        retval = dataset.output()

        pos_labels = self._pos_labels.copy()  # take care not to update the base values.
        pos_labels.update(dataset.metadata.pos_labels)

        retval_data = dataset.get_dataframe()
        for role in pos_labels:
            if role not in dataset.roles.keys():
                continue

            columns_satisfying_role = dataset.roles[role]

            positive = pos_labels[role]
            for col in columns_satisfying_role:
                # json lacks a bytes representation. We use str.
                first = retval_data[col.name].iloc[0]
                if isinstance(first, bytes) and isinstance(positive, str):
                    positive = positive.encode()
                logger.debug('positive: %r', positive)
                retval_data[col.name] = (
                    retval_data[col.name].apply(lambda x, pos=positive: x == pos))

        retval.dataframe = retval_data

        return retval


class BinarizerModel(Algorithm):
    '''Convert two-valued fields into binary fields.

    Takes a DatasetKeys.DATAFRAME.
    Yields a DatasetKeys.DATAFRAME.

    A column needs to have a pos_label property in the
    problem definition to be converted.

    If the field is already boolean, no change is made.

    If the field is bytes, the pos_label will be converted
    to bytes for the default encoding with str.encode().
    '''
    _name = 'binarize'
    _tags = {
        'source': ['auton_lab'],
        'preprocessor': ['true'],
    }
    _instance_constructor = BinarizerModelInstance


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = BinarizerModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Tests for k_fold_splitter.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.
from typing import List

import pandas as pd

from ..problem_def.cross_validation_config import CrossValidationConfig
from ..wrangler.dataset import Dataset, Metadata, Column, RoleName

from .k_fold_splitter import KFoldSplitter

# pylint: disable=missing-function-docstring, duplicate-code


def test_split_default_hyperparams() -> None:
    dataframe = pd.DataFrame({
        'a': range(1, 1001),
        'b': range(1001, 2001),
        'c': range(2001, 3001)})
    data = Dataset(metadata=Metadata(
        roles={
            RoleName.TARGET: [Column('c')]
        }
    ))
    data.dataframe = dataframe
    dut = KFoldSplitter(cv_config=CrossValidationConfig({}))

    got = dut.split(dataset=data)

    # default n_splits: 5
    assert len(got.folds) == 5
    assert got.ground_truth is not None
    assert got.ground_truth.ground_truth.shape == (1000, 1)
    for fold in got.folds:
        # every row not in the fold should be in the train dataset
        # all three columns are in the train dataset
        assert fold.train.dataframe.shape == (800, 3)
        # total number of rows in each fold is 200
        # validate dataset does not include the target column
        assert fold.validate.dataframe.shape == (200, 2)


def test_split_is_random() -> None:
    dataframe = pd.DataFrame({
        'a': range(1, 1001),
        'b': range(1001, 2001),
        'c': range(2001, 3001)})
    data = Dataset(metadata=Metadata(
        roles={
            RoleName.TARGET: [Column('c')]
        }
    ))

    data.dataframe = dataframe
    dut = KFoldSplitter(cv_config=CrossValidationConfig({}))

    got = dut.split(dataset=data)
    assert got.folds[0].train.dataframe.iat[10, 1] == 1014

    dut2 = KFoldSplitter(cv_config=CrossValidationConfig({'seed': 1054}))
    got2 = dut2.split(dataset=data)
    assert got2.folds[0].train.dataframe.iat[10, 1] == 1015


def test_split_large_dataset() -> None:
    dataframe = pd.DataFrame(
        {
            'a': range(1, 15001),
            'b': range(1, 15001),
            'c': range(1, 15001)
        }
    )
    data = Dataset(metadata=Metadata(
        roles={
            RoleName.TARGET: [Column('c')]
        }
    ))

    data.dataframe = dataframe
    dut = KFoldSplitter(cv_config=CrossValidationConfig({}))
    got = dut.split(dataset=data)

    assert len(got.folds) == 3
    assert got.ground_truth is not None
    assert got.ground_truth.ground_truth.shape == (15000, 1)

    for fold in got.folds:
        # every row not in the fold should be in the train dataset
        # all three columns are in the train dataset
        assert fold.train.dataframe.shape == (10000, 3)

        # total number of rows in each fold is 5,000
        # validate dataset does not include the target column
        assert fold.validate.dataframe.shape == (5000, 2)


def test_split_order() -> None:
    dataframe = pd.DataFrame({
        'a': range(1, 101),
        'b': range(101, 201),
        'c': range(201, 301)})
    data = Dataset(metadata=Metadata(
        roles={
            RoleName.TARGET: [Column('c')]
        }
    ))

    data.dataframe = dataframe
    dut = KFoldSplitter(cv_config=CrossValidationConfig({}))
    got = dut.split(dataset=data)
    validate_dfs: List[pd.DataFrame] = [f.validate.dataframe for f in got.folds]
    concatenated_folds = pd.concat(validate_dfs, axis=0)
    concatenated_folds.reset_index(inplace=True, drop=True)
    assert concatenated_folds.shape == (100, 2)
    assert got.ground_truth is not None
    assert got.ground_truth.ground_truth is not None
    list_to_concat: List[pd.DataFrame] = [concatenated_folds, got.ground_truth.ground_truth]
    full_dataset = pd.concat(list_to_concat, axis=1)
    assert full_dataset.shape == (100, 3)
    assert all(full_dataset['a'] + 200 == full_dataset['c'])



'''Autoloader for the model catalog.'''
# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path

from ...catalog.memory_catalog import MemoryCatalog
from ...catalog.plugin_catalog import PluginCatalog
from .algorithm import AlgorithmCatalog, Algorithm


class AlgorithmCatalogAuto(PluginCatalog[Algorithm], AlgorithmCatalog):
    '''Model catalog that Automatically loads all models from a directory.

    The models are in ../algorithms.
    '''

    def __init__(self, **kwargs):
        super().__init__(catalog_name='algorithms', default_root=Path(__file__).parents[2])


class FakeAlgorithmCatalogAuto(MemoryCatalog[Algorithm], AlgorithmCatalog):
    '''Algorithm catalog that loads only a subset of algorithms.

    Namely, those that have the {'for_tests': ['true']} tag.
    The goal is to streamline the tests.
    '''

    def __init__(self, **kwargs):
        super().__init__()
        catalog = AlgorithmCatalogAuto(**kwargs)
        for name, algorithm in catalog.lookup_by_tag_and(for_tests="true").items():
            self.register(algorithm, name=name)



'''Base class for sklearn data loaders'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code,too-many-arguments
import abc
from typing import Any, Optional

import pandas as pd

from ..impl.data_loader_catalog import DataLoaderCatalog
from ..impl.dataframe_loader import DataframeLoader


class SklearnLoaderInterface(DataframeLoader, metaclass=abc.ABCMeta):
    '''Generates constructors for SklearnDataLoaders.'''
    _impl: Any
    _test_data: Optional[pd.DataFrame] = None


def register(catalog: DataLoaderCatalog):  # pylint: disable=unused-argument
    '''Nothing to register.

    All subclasses of SklearnAlgorithm are registered in sklearn_algorithms.py
    '''



'''Utility functions for managing DistributedAlgorithm synchronous mode.'''
# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import time
from typing import Optional, Sequence

from .distributed_algorithm_instance import DistributedAlgorithmInstance


class NotSynchronousError(Exception):
    '''The method is not available in asynchronous mode.'''


def fit_now(duts: Sequence[DistributedAlgorithmInstance]) -> None:
    '''Fit the models now.

    This is a convenience function to fit all the models in a list of DistributedAlgorithmInstances.
    '''
    for d in duts:
        if not d.synchronous:
            raise NotSynchronousError(
                f'Node {d.my_id} (algorithm {d.catalog_name}) is attempted synchronous fit_now,'
                ' but is not running synchronously.')
        d.fit_now()


def read_from_neighbors(duts: Sequence[DistributedAlgorithmInstance],
                        min_time: int = 100_000_000,
                        timeout_ns: Optional[int] = None) -> None:
    '''Wait for pending messages from neighbors to be processed.

    We wait for at least min_time nanoseconds (default 0.1 sec), and at most
    timeout_ns nanoseconds. If timeout_ns is None (default), we wait indefinitely.
    '''
    _ = timeout_ns  # BUG: We need to extend the Communicator API to support this.
    time.sleep(min_time / 1_000_000_000)  # Allow for propagation delay.
    for d in duts:
        # BUG: We could theoretically wait len(duts) * timeout_ns, but we expect
        # that messages should mostly get processed during the first delay and
        # that subsequent calls will be much faster.
        if not d.synchronous:
            raise NotSynchronousError(
                f'Node {d.my_id} (algorithm {d.catalog_name}) is attempted synchronous read,'
                ' but is not running synchronously.')
        d.read_from_neighbors()


def advance(duts: Sequence[DistributedAlgorithmInstance],
            min_time: int = 100_000_000,
            timeout_ns: Optional[int] = None) -> None:
    '''Advance the models by one step.

    This is a convenience function to advance all the models in a list of
    DistributedAlgorithmInstances.
    '''
    read_from_neighbors(duts=duts, min_time=min_time, timeout_ns=timeout_ns)
    fit_now(duts=duts)



'''Tests for boolan.py and related files.'''
# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Mapping

import pytest

from ..algorithms.impl.neighbor_metadata import NeighborMetadata
from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..algorithms.impl.toy_neighbor_state import ToyNeighborState
from ..config_components.distributed_config import DistributedConfig
from ..deciders.impl.decider import Decider
from ..neighbor_manager.node_id import NodeID

from .impl.decider_auto import DeciderCatalogAuto


@pytest.mark.parametrize('op,lhs,rhs,result', [
    ('and', 'true', 'true', True),
    ('and', 'true', 'false', False),
    ('and', 'false', 'true', False),
    ('and', 'false', 'false', False),
    ('or', 'true', 'true', True),
    ('or', 'true', 'false', True),
    ('or', 'false', 'true', True),
    ('or', 'false', 'false', False),
    ('not', 'true', 'true', False),
    ('not', 'false', 'false', True),
])
def test_and_or(op: str, lhs: str, rhs: str, result: bool):
    '''Test the decider_stub in the DeciderCatalog class.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                op: {
                    lhs: {},
                    rhs: {}
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name(op)(config=getattr(distributed.decider, op))  # pylint: disable=no-member

    my_meta = NeighborMetadata(my_state=ToyNeighborState(center=0.0))
    neighbor_id = NodeID(1)
    neighbors = {
        neighbor_id: NeighborMetadata(my_state=ToyNeighborState(center=0.0))
    }

    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) == result


@pytest.mark.parametrize('threshold,result', [
    (0.1, False),
    (0.5, True),
])
def test_compound(threshold: float, result: bool):
    '''Test a compund Boolean decider.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'and': {
                    'true': {},
                    'or': {
                        'not': {
                            'distance': {
                                'threshold': threshold,
                            },
                        },
                        'false': {},
                    }
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('and')(config=getattr(distributed.decider, 'and'))  # pylint: disable=no-member

    my_meta = NeighborMetadata(my_state=ToyNeighborState(center=0.0))
    neighbor_id = NodeID(1)
    neighbors: Mapping[NodeID, NeighborMetadataInterface] = {
        neighbor_id: NeighborMetadata(my_state=ToyNeighborState(center=0.5))
    }

    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors) is result



'''Tests for k_fold_cross_validator.py'''
# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring, missing-class-docstring
# pylint: disable=redefined-outer-name,duplicate-code
from typing import Dict, List

import pandas as pd
import pytest


from ..executor.executor import ExecutorStub
from ..generator.bound_pipeline import BoundPipeline, BoundPipelineStub
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                PipelineResult,
                                                PredictError)
from ..instantiator.instantiator_factory import InstantiatorFactory
from ..splitters.impl.splitter import Fold, SplitDataset
from ..tables.impl.table import TableFactory
from ..tables.impl.table_auto import TableCatalogAuto
from ..wrangler.dataset import Dataset, Metadata, RoleName, Column
from .k_fold_cross_validator import KFoldCrossValidator

_ = TableCatalogAuto()  # pylint: disable=pointless-statement


class FakeExecutor(ExecutorStub):
    _return_proba: bool

    def __init__(self,
                 return_proba: bool = False,
                 error: bool = False,
                 order_test: bool = False,
                 **kwargs) -> None:
        self._return_proba = return_proba
        self._error = error
        self._order_test = order_test
        super().__init__(**kwargs)

    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Use a list of pipelines to predict from a dataset.'''

        predictions = dataset.output()
        predictions.predictions = dataset.dataframe

        if self._order_test:
            df = pd.DataFrame({'target': dataset.dataframe['other_column'] + 1})
            predictions.predictions = df

        if self._return_proba:
            probabilities_df = pd.DataFrame(
                {'prob': [0.5, 0.6, 0.7],
                    'id': [0, 1, 2]})
            probabilities_df.set_index('id', inplace=True)
            predictions.probabilities = TableFactory(probabilities_df)  # type: ignore[abstract] # pylint: disable=abstract-class-instantiated
        retval = {d: PipelineResult(executable_pipeline=p, prediction=predictions)
                  for d, p in pipelines.items()}

        if self._error:
            error_predictions = predictions.output()
            error_predictions['error'] = PredictError('there was an error')
            first = sorted(list(retval.keys()))[0]
            retval[first] = PipelineResult(
                executable_pipeline=pipelines[first],
                prediction=error_predictions)

        return retval


@pytest.fixture(scope='session')
def split_dataset() -> SplitDataset:
    fake_dataset = Dataset()
    fake_dataset.dataframe = pd.DataFrame(
        {'a': [53, 4, 5030],
         'id': [0, 1, 2]}).set_index('id')
    folds = [
        Fold(train=fake_dataset, validate=fake_dataset),
        Fold(train=fake_dataset, validate=fake_dataset)]
    return SplitDataset(folds)


def test_sunny_day(split_dataset: SplitDataset) -> None:
    bp1 = BoundPipelineStub('bound_pipeline_stub_1')
    bp2 = BoundPipelineStub('bound_pipeline_stub_2')
    des1 = Designator('des1')
    des2 = Designator('des2')
    bound_pipelines: Dict[Designator, BoundPipeline] = {
        des1: bp1, des2: bp2}

    dut = KFoldCrossValidator()
    got = dut.validate_pipelines(
        split_dataset=split_dataset,
        bound_pipelines=bound_pipelines,
        instantiator=InstantiatorFactory(),
        executor=FakeExecutor(return_proba=False))

    assert des1 in got and des2 in got
    assert len(got) == 2
    assert got[des1].bound_pipeline.name == bp1.name
    assert got[des2].bound_pipeline.name == bp2.name
    assert got[des1].split_dataset == split_dataset
    assert got[des2].split_dataset == split_dataset
    expected_frame = pd.DataFrame({
        'a': [53, 4, 5030, 53, 4, 5030],
        'id': [0, 1, 2, 0, 1, 2]}).set_index('id')

    got_des1 = got[des1]
    got_des2 = got[des2]
    assert got_des1.prediction is not None
    assert got_des2.prediction is not None

    pd.testing.assert_frame_equal(
        got_des1.prediction.predictions, expected_frame)
    pd.testing.assert_frame_equal(
        got_des2.prediction.predictions, expected_frame)


def test_probabilities_preserved(
        split_dataset: SplitDataset) -> None:

    bp1 = BoundPipelineStub('bound_pipeline_stub_1')
    bp2 = BoundPipelineStub('bound_pipeline_stub_2')
    des1 = Designator('des1')
    des2 = Designator('des2')
    bound_pipelines: Dict[Designator, BoundPipeline] = {
        des1: bp1, des2: bp2}

    dut = KFoldCrossValidator()
    got = dut.validate_pipelines(
        split_dataset=split_dataset,
        bound_pipelines=bound_pipelines,
        instantiator=InstantiatorFactory(),
        executor=FakeExecutor(return_proba=True))

    expected_frame = pd.DataFrame({
        'prob': [0.5, 0.6, 0.7, 0.5, 0.6, 0.7],
        'id': [0, 1, 2, 0, 1, 2]}).set_index('id')

    got_des1 = got[des1]
    got_des2 = got[des2]
    assert got_des1.prediction is not None
    assert got_des2.prediction is not None

    pd.testing.assert_frame_equal(
        got_des1.prediction.probabilities.as_(pd.DataFrame),  # type: ignore[attr-defined]
        expected_frame)
    pd.testing.assert_frame_equal(
        got_des2.prediction.probabilities.as_(pd.DataFrame), expected_frame)  # type: ignore[attr-defined] # pylint: disable=line-too-long


def test_error_catching(split_dataset: SplitDataset) -> None:
    '''Properly handle pipeline errors that are caught by the executor

    We expect to find no 'predictions' key in the dataset, but an 'errors' key instead
    '''
    bp1 = BoundPipelineStub('bound_pipeline_stub_1')
    bp2 = BoundPipelineStub('bound_pipeline_stub_2')
    des1 = Designator('des1')
    des2 = Designator('des2')
    bound_pipelines: Dict[Designator, BoundPipeline] = {
        des1: bp1, des2: bp2}

    dut = KFoldCrossValidator()
    got = dut.validate_pipelines(
        split_dataset=split_dataset,
        bound_pipelines=bound_pipelines,
        instantiator=InstantiatorFactory(),
        executor=FakeExecutor(error=True))

    assert des1 in got and des2 in got
    assert len(got) == 2
    assert got[des1].bound_pipeline.name == bp1.name
    assert got[des2].bound_pipeline.name == bp2.name
    assert got[des1].split_dataset == split_dataset
    assert got[des2].split_dataset == split_dataset

    got_des1 = got[des1]
    assert got_des1.prediction is not None
    assert 'predictions' not in got_des1.prediction
    assert isinstance(got_des1.prediction['error'], PredictError)

    got_des2 = got[des2]
    assert got_des2.prediction is not None
    assert 'error' not in got_des2.prediction
    expected_frame = pd.DataFrame({
        'a': [53, 4, 5030, 53, 4, 5030],
        'id': [0, 1, 2, 0, 1, 2]}).set_index('id')
    pd.testing.assert_frame_equal(
        got_des2.prediction.predictions, expected_frame)


def test_order() -> None:
    bp1 = BoundPipelineStub('bound_pipeline_stub_1')
    des1 = Designator('des1')

    bound_pipelines: Dict[Designator, BoundPipeline] = {des1: bp1}

    met = Metadata(roles={RoleName.TARGET: [Column('target')]})
    empty_dataset = Dataset(metadata=met)
    folds: List[Fold] = []
    for vals in [[1, 2, 3], [4, 5, 6], [7, 8, 9]]:
        folds.append(Fold(
            train=empty_dataset,
            validate=Dataset(
                dataframe=pd.DataFrame(
                    {'other_column': vals}
                ),
                metadata=met
            )
        ))

    split_dataset = SplitDataset(folds)

    dut = KFoldCrossValidator()
    got = dut.validate_pipelines(
        split_dataset=split_dataset,
        bound_pipelines=bound_pipelines,
        instantiator=InstantiatorFactory(),
        executor=FakeExecutor(order_test=True))

    result = got[des1]
    assert result.prediction is not None
    print(result.prediction.predictions)
    assert list(result.prediction.predictions['target']) == [2, 3, 4, 5, 6, 7, 8, 9, 10]



'''Tests for queue_loader.py.'''

import queue

import pandas as pd

from ..config_components.dataset_config import DatasetConfig
from ..problem_def.problem_def import ProblemDefinition
from ..wrangler.dataset import Dataset

from .queue_loader import QueueLoader

# pylint: disable=missing-function-docstring, missing-class-docstring, duplicate-code


def test_sunny_day() -> None:
    '''Test the happy path for the QueueLoader.'''
    clause = {
        'config': 'queue',
        'column_roles': {
            'target': {
                'name': 'a',
                'pos_label': 1,
            },
        },
        'params': {
            'train_data': 'train_df',
            'test_data': 'test_df',
            'poll_queue': 'poll_queue',
            'timeout': 0.0,
        },
    }
    config = DatasetConfig(clause=clause)

    train_df = pd.DataFrame({'a': [1, 2], 'b': [3, 4]}) # noqa F841 pylint: disable=unused-variable
    test_df = pd.DataFrame({'a': [5, 6], 'b': [7, 8]}) # noqa F841 pylint: disable=unused-variable
    poll_data = Dataset(
        dataframe=pd.DataFrame({'a': [9, 10], 'b': [11, 12]}),
        metadata=config.metadata)

    poll_queue: queue.Queue[Dataset] = queue.Queue()

    dut = QueueLoader(config=config)

    assert dut.poll(timeout=0) is None

    poll_queue.put(poll_data)

    got_train = dut.load_train()
    got_test = dut.load_test()
    got_poll = dut.poll(timeout=0)

    assert got_train is not None
    assert got_test is not None
    assert got_poll is not None

    want_train = pd.DataFrame({'a': [1, 2], 'b': [3, 4]})
    want_test = pd.DataFrame({'b': [7, 8]})
    want_poll = poll_data.dataframe

    pd.testing.assert_frame_equal(got_train.dataframe, want_train)
    pd.testing.assert_frame_equal(got_test.dataframe, want_test)
    pd.testing.assert_frame_equal(got_poll.dataframe, want_poll)


def test_split() -> None:
    '''Allow for a split dataset.'''
    pdef = ProblemDefinition({
        'dataset': {
            'config': 'queue',
            'column_roles': {
                'target': {
                    'name': 'a',
                    'pos_label': 1,
                },
            },
            'params': {
                'train_data': 'train_df',
                'test_data': 'test_df',
                'poll_queue': 'poll_queue',
                'timeout': 0.0,
            },
        },
        'distributed': {
            'my_id': 1,
            'split': {
                'num_nodes': 3,
                'seed': 1234
            },
        },
        'problem_type': {
            'task': 'regression'
        }
    })
    config = pdef.get_conf('dataset')
    assert isinstance(config, DatasetConfig)
    df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9],
    })

    train_df = df # noqa F841 pylint: disable=unused-variable
    test_df = pd.DataFrame({'a': [10, 11, 12], 'b': [13, 14, 15], 'c': [16, 17, 18]}) # noqa F841 pylint: disable=unused-variable
    poll_data1 = Dataset(
        dataframe=pd.DataFrame({'a': [19, 20, 21], 'b': [22, 23, 24], 'c': [25, 26, 27]}),
        metadata=config.metadata)
    poll_data2 = Dataset(
        dataframe=pd.DataFrame({'a': [28, 29, 30], 'b': [31, 32, 33], 'c': [34, 35, 36]}),
        metadata=config.metadata)

    poll_queue: queue.Queue[Dataset] = queue.Queue()

    dut = QueueLoader(config=config)

    assert dut.poll(timeout=0) is None

    poll_queue.put(poll_data1)
    poll_queue.put(poll_data2)

    want_train = pd.DataFrame({'a': [1], 'b': [4], 'c': [7]})
    # Test data shouldn't be split, but it should have the target removed.
    want_test = pd.DataFrame({'b': [13, 14, 15], 'c': [16, 17, 18]})
    want_poll1 = pd.DataFrame({'a': [19], 'b': [22], 'c': [25]})
    want_poll2 = pd.DataFrame({'a': [28], 'b': [31], 'c': [34]})

    got_train = dut.load_train()
    got_test = dut.load_test()
    got_poll1 = dut.poll(timeout=0)
    got_poll2 = dut.poll(timeout=0)

    assert got_train is not None
    assert got_test is not None
    assert got_poll1 is not None
    assert got_poll2 is not None

    pd.testing.assert_frame_equal(got_train.dataframe, want_train)
    pd.testing.assert_frame_equal(got_test.dataframe, want_test)
    pd.testing.assert_frame_equal(got_poll1.dataframe, want_poll1)
    pd.testing.assert_frame_equal(got_poll2.dataframe, want_poll2)



'''Test the config component catalog'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, List
from .config_component import ConfigComponent
from .config_component_catalog import ConfigComponentCatalogStub
# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code


class TempConfigComponent(ConfigComponent):

    name = "temp_config_component"
    tags: Dict[str, List[str]] = {}

    def validate(self, **kwargs) -> None:
        pass


def test_register() -> None:
    dut = ConfigComponentCatalogStub()
    assert 'temp_config_component' == dut.register(obj=TempConfigComponent)


def test_lookup_by_name() -> None:
    dut = ConfigComponentCatalogStub()
    some_component = TempConfigComponent
    dut.register(obj=some_component)
    got = dut.lookup_by_name('temp_config_component')
    assert isinstance(got, type)
    assert isinstance(got(clause={}), TempConfigComponent)



'''Auton Lab's implementation of a decision tree classifier for distributed data.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code,too-many-lines

import copy
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from ...catalog.catalog import upcast
from ...config_components.distributed_config import DistributedConfig
from ...problem_def.task import DataType, TaskType
from ...tables.impl.table import TableFactory
from ...wrangler.constants import Defaults
from ...wrangler.dataset import Dataset, DatasetKeys
from ...wrangler.logger import Logger, Level

from ..impl.distributed_algorithm_instance import DistributedAlgorithmInstance, NeighborState
from ..impl.algorithm import Algorithm, HyperparamError, MemoryAlgorithmCatalog

from .auton_decision_tree_neighbor import AutonDecisionTreeNeighbor

log = Logger(__file__, to_stdout=False,
             level=Level.INFO).logger()


def loss_gini(v: np.ndarray) -> float:
    '''Standard gini loss function'''
    return 1 - (v ** 2).sum(axis=-1)


class AutonDecisionTreeInstance(DistributedAlgorithmInstance):
    '''Implementation of Gaussian Mixture Models for distributed data.'''
    _neighbor_constructor = AutonDecisionTreeNeighbor

    _predict_state: Optional[AutonDecisionTreeNeighbor] = None
    _max_depth: Optional[int] = None
    _min_sample_weight: float = 0.0
    _min_vol: float = 0.0
    _ccp_alpha: float = 0.0
    _loss_fn: Callable[[np.ndarray], float]

    _random_state: np.random.RandomState  # pylint: disable=no-member,line-too-long

    def __init__(self,
                 parent: Algorithm,
                 distributed: DistributedConfig,
                 **kwargs):
        hyperparams = parent.hyperparams(**kwargs)

        _random_seed = hyperparams.pop('random_seed', None)
        self._random_state = np.random.RandomState(_random_seed)  # pylint: disable=no-member,line-too-long

        if not isinstance(distributed, DistributedConfig):
            raise HyperparamError(
                f'{self.__class__.__name__} distributed must be a DistributedConfig, instead found '
                f'{distributed} of type {type(distributed)}')

        max_depth = hyperparams.pop('max_depth')
        self._max_depth = None if max_depth is None else int(max_depth)
        self._min_sample_weight = float(hyperparams.pop('min_sample_weight'))
        self._min_vol = float(hyperparams.pop('min_vol'))
        self._ccp_alpha = float(hyperparams.pop('ccp_alpha'))
        self._loss_fn = hyperparams.pop('loss_fn')
        # Either max_depth, min_sample_weight, min_vol or ccp_alpha should be set or trees can
        #   grow very large.
        # TODO(Merritt): Is there a way to solve this without a warning?
        if self._max_depth is None and self._min_sample_weight == 0.0 and (
            self._min_vol == 0.0 and (
                self._ccp_alpha == 0.0)):
            log.warning(
                'No stopping condition set so nothing to limit tree growth. '
                'Consider setting max_depth, min_sample_weight, '
                'min_vol or ccp_alpha.')

        # TODO(Merritt): allow setting lower and upper bound & number of classes via hyperparam

        super().__init__(parent, distributed=distributed, **hyperparams)

    @property
    def _neighbor_models_iter(self) -> Iterator[AutonDecisionTreeNeighbor]:
        for v in super()._neighbor_models_iter:
            assert isinstance(v, AutonDecisionTreeNeighbor), (
                'BUG: expected neighbor_models to contain AutonDecisionTreeNeighbor, '
                f'instead found {v} of type {type(v)}'
            )
            yield v

    @property  # type: ignore[override]
    def _my_state(self) -> Optional[AutonDecisionTreeNeighbor]:
        retval = DistributedAlgorithmInstance._my_state.fget(self)  # type: ignore[attr-defined] # pylint: disable=assignment-from-no-return,line-too-long
        assert retval is None or isinstance(retval, AutonDecisionTreeNeighbor), (
            'BUG: expected _my_state to be None or an AutonDecisionTreeNeighbor.')
        return retval

    @_my_state.setter
    def _my_state(self, value: Optional[AutonDecisionTreeNeighbor]) -> None:
        assert value is None or isinstance(value, AutonDecisionTreeNeighbor), (
            'BUG: expected value to be None or an AutonDecisionTreeNeighbor.')
        DistributedAlgorithmInstance._my_state.fset(self, value)  # type: ignore[attr-defined]

    def _decode(self, serialized_model: bytes) -> NeighborState:
        '''Decode a message from distributed neighbors. '''
        return AutonDecisionTreeNeighbor.decode(serialized_model)

    def _is_leaf(self, i: int) -> bool:
        '''Is node i a leaf?'''
        assert self._predict_state is not None, (
            'BUG: _predict_state must exist when calling is_leaf.')
        return self._predict_state.is_leaf(i)

    def _internal_node_info(self, i: int) -> Tuple[int, float, bool, int, int]:
        '''Return (feature, threshold, equal, children_left, children_right)

        while asserting that this node is not a leaf.
        '''
        assert self._predict_state is not None, (
            'BUG: _predict_state must not be None when calling internal_node_info'
        )
        return self._predict_state.internal_node_info(i)

    def _reconcile_class_list(self,
                              data_classes: Optional[List[str]],
                              prev_fit_class_list: Optional[List[str]]) -> List[str]:
        '''Reconcile recorded class list with neighbor class list and classes in the data.

        data_classes: classes observed in the data, or None if fitting with no data
        class_list: recorded class list from previous fit, or None if first fit

        Returns new class list.
        '''

        trees = list(self._neighbor_models_iter)
        some_neighbor_class_list = None if len(trees) == 0 else trees[0].class_list

        assert any([data_classes, prev_fit_class_list, some_neighbor_class_list])

        all_class_lists: List[List[str]] = []
        if prev_fit_class_list is not None:
            all_class_lists.append(prev_fit_class_list)
        if data_classes is not None:
            data_classes = sorted(data_classes)
            all_class_lists.append(data_classes)
        if len(trees) > 0:
            all_class_lists += [m.class_list for m in trees]

        # if all class lists match: nothing to do
        assert len(all_class_lists) > 0
        some_class_list = all_class_lists[0]
        if all(cl == some_class_list for cl in all_class_lists):
            return some_class_list

        new_class_list = sorted(set().union(*all_class_lists))
        new_n_out = len(new_class_list)

        log.log(Level.VERBOSE,
                'Prev fit class list: %s\nData class list: %s\n'
                'Neighbor class lists: %s\nNew class list: %s',
                prev_fit_class_list, data_classes, [m.class_list for m in trees], new_class_list)

        # modify trees in place
        def permute_values(t: Optional[AutonDecisionTreeNeighbor],
                           to_: np.ndarray, from_: np.ndarray, new_n_out: int) -> None:
            if t is None:
                return
            for i, old_val in enumerate(t.value):
                if old_val is None:
                    log.debug('Index %d skipping None old_val', i)
                    continue
                v = np.zeros(new_n_out)

                # from_ represents indices in old_val and to_ represents corresponding indices in v
                assert len(to_) == len(from_)
                assert all(i < len(v) for i in to_)
                assert all(i < len(old_val) for i in from_)
                v[to_] = old_val[from_]
                t.value[i] = v

        for i, m in enumerate(trees):
            _, to_, from_ = np.intersect1d(
                new_class_list, m.class_list, assume_unique=True, return_indices=True)
            try:
                permute_values(m, to_, from_, new_n_out)
                m.modify_classes(class_list=new_class_list)
            except AssertionError as err:
                msg = ('Error while reconciling class list has been '
                       f'encountered on tree {i}.\n'
                       f'Prev fit class list: {prev_fit_class_list}\n'
                       f'Data class list: {data_classes}\n',
                       f'Neighbor class lists: {[m.class_list for m in trees]}\n'
                       f'New class list: {new_class_list}\n')
                raise AssertionError(msg) from err
            # TODO(Merritt): is it fine to modify local representations of neighbors?  talk to kyle

        # modify self
        if self._my_state is not None:
            old_class_list = self._my_state.class_list
            _, to_, from_ = np.intersect1d(
                new_class_list, old_class_list, assume_unique=True, return_indices=True)
            permute_values(self.my_state, to_, from_, new_n_out)
            self._my_state.modify_classes(class_list=new_class_list)

        return new_class_list

    def _append_leaf(self, v: np.ndarray, w: float) -> None:
        '''Modify our state in-place to append a leaf.'''
        assert self._my_state is not None, (
            'BUG: cannot append leaf before state is initialized.')
        self._my_state.children_left.append(None)
        self._my_state.children_right.append(None)
        self._my_state.feature.append(None)
        self._my_state.threshold.append(None)
        self._my_state.equal.append(None)
        self._my_state.value.append(v)
        self._my_state.sample_weight.append(w)

    def _prepare_data(self,
                      dataset: Optional[Dataset]) -> Tuple[
                          np.float64, np.ndarray, np.ndarray, List[str]]:
        if dataset is None:
            assert self._neighbor_metadata, (
                'BUG: expected either dataset or neighbor metadata to be non-None'
            )
            # No data so we only synthesize neighbor models
            # Set omega to 1 (no self-regularization)
            _omega = np.float64(1.0)
            # Create blank data with 0 rows and a number of columns matching neighbors
            ncol = next(self._neighbor_models_iter).n_in
            cols = next(self._neighbor_models_iter).columns
            assert cols is not None, (
                'BUG: found neighbor model with no recored columns'
            )
            if any(model.n_in != ncol for model in self._neighbor_models_iter):
                raise NotImplementedError(
                    'Neighbor models disagree about number of columns '
                    'and reconciling cols is not implemented.')
            x = np.zeros(shape=(0, ncol))
            y = np.zeros(shape=(0, 1))
            log.debug("preparing data for empty dataset: _omega: %s, x: %s, y: %s, cols: %s",
                      _omega, x, y, cols)
        else:
            _omega = self._omega
            target = dataset.metadata.target
            assert target is not None, 'BUG: found no target col'
            if DatasetKeys.DATAFRAME.value in dataset:
                orig_cov = dataset.dataframe.copy()
                cols = list(orig_cov.columns)
                orig_cov.drop(columns=[target.name], inplace=True)
                orig_tar = dataset.dataframe.loc[:, target.name]
                x = orig_cov.to_numpy()
                y = orig_tar.to_numpy().flatten()  # type: ignore[assignment]
            else:
                cov_df = dataset[DatasetKeys.COVARIATES.value]
                assert isinstance(cov_df, pd.DataFrame), (
                    f'BUG: expected dataframe, instead found {cov_df} of type {type(cov_df)}'
                )
                x = cov_df.to_numpy()
                cols = list(cov_df.columns)
                y = dataset[DatasetKeys.TARGET.value][target.name].to_numpy().flatten()
            log.debug("preparing data with non-empty dataset: _omega: %s, x: %s, y: %s, cols: %s",
                      _omega, x, y, cols)

        return _omega, x, y, cols

    def _reconcile_bounds(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        lower_bound_candidates: List[np.ndarray] = []
        upper_bound_candidates: List[np.ndarray] = []
        if x.shape[0] > 0:
            # we have data
            lower_bound_candidates.append(x.min(0))
            upper_bound_candidates.append(x.max(0))
        if self._my_state is not None:
            # we have a previous fit
            lower_bound_candidates.append(self._my_state.lower_bound)
            upper_bound_candidates.append(self._my_state.upper_bound)
        trees = list(self._neighbor_models_iter)
        lower_bound_candidates += [m.lower_bound for m in trees]
        upper_bound_candidates += [m.upper_bound for m in trees]
        lower_bound: np.ndarray = np.stack(lower_bound_candidates).min(0)
        upper_bound: np.ndarray = np.stack(upper_bound_candidates).max(0)
        return (lower_bound, upper_bound)

    def _fit(self,  # pylint: disable=too-many-statements,too-many-locals
             dataset: Optional[Dataset],
             **kwargs) -> None:
        '''Fit the model to the data. This is the actual implementation of fit.'''
        if dataset is None and not self._neighbor_metadata:
            # No data or neighbor models; cannot fit
            return
        log.debug("fit on dataset: %s \n self._neighbor_metadata: %s \n",
                  dataset, self._neighbor_metadata)
        _omega, x, y, cols = self._prepare_data(dataset=dataset)
        lower_bound, upper_bound = self._reconcile_bounds(x=x)

        # reconcile class list
        old_class_list = None if self.my_state is None else self.my_state.class_list
        if y.shape[0] == 0:
            class_list = self._reconcile_class_list(
                data_classes=None,
                prev_fit_class_list=old_class_list
            )
            new_n_out = len(class_list)
        elif y.ndim == 1:
            # y is 1 column: format y as one-hot labels
            data_classes = np.unique(y).tolist()
            class_list = self._reconcile_class_list(
                data_classes=data_classes, prev_fit_class_list=old_class_list)
            new_n_out = len(class_list)
            y1hot = np.zeros((x.shape[0], new_n_out))
            for i, c in enumerate(class_list):
                y1hot[y == c, i] = 1
            y = y1hot.reshape((*y.shape, new_n_out))
        else:
            # y is already formatted as one-hot labels
            # HACK assume one-hot columns correspond to class-ids (naive)
            data_classes = np.arange(y.shape[1]).tolist()
            class_list = self._reconcile_class_list(
                data_classes=data_classes, prev_fit_class_list=class_list)
            new_n_out = len(class_list)

        # prepare other trees
        trees = list(self._neighbor_models_iter)
        if self._my_state is not None:
            trees = [self._my_state] + trees

            tree_weights = self._lambda * np.ones(len(trees))
            tree_weights[0] *= (1 - _omega) / _omega
        else:
            tree_weights = self._lambda * np.ones(len(trees))

        for i, tree in enumerate(trees):
            assert tree.num_nodes > 0, (
                f'Tree {i} is empty: {tree}'
            )
            assert len(tree.value[0]) == len(tree.class_list), (
                f'Tree {i} has values mismatched with its class list. \n'
                f'Values: {tree.value} \n'
                f'Class list: {tree.class_list} \n'
                f'My state: {self.my_state}\n'
            )
            assert tree.class_list == trees[0].class_list

        n_in = x.shape[1]
        self._my_state = AutonDecisionTreeNeighbor.init_tree(
            n_in, new_n_out, lower_bound, upper_bound, class_list,
            columns=cols
        )

        def recursive_fit(d: int,  # pylint: disable=too-many-statements,too-many-locals,too-many-branches
                          x: np.ndarray,
                          y: np.ndarray,
                          lb: np.ndarray,
                          ub: np.ndarray,
                          n: float) -> int:
            '''
            d: depth of this node  [varies]
            x: local data (covariates)  [constant]
            y: local data (target) [constant]
            lb: lower bound of this subtree in the input space [varies]
            up: upper bound of this subtree in the input space [varies]
            n: number of samples plus total weight of all other trees; [constant]
                used to normalize the loss for subtrees

            Returns: this node's index

            notation in this section
            s: the sum of label weights for each class in a given region
            w: the weight of samples and other trees' influence in a given region
            v: the distribution of labels in a given region, v = s/w
            t: the threshold value for splitting
            eq: flag to mark a threshold as x <= t, as opposed to x < t
            ds: the instantaneous change in s at a given threshold
            dsdt: the rate of change of s with respect to t in a given region
            ddsdt: the instantaneous change in dsdt at a given threshold
            {x}L: the value of some variable {x} for the left child
            {x}R: the value of some variable {x} for the right child
            '''
            log.debug('calling recursive fit, d=%s', d)
            assert self._my_state is not None

            # get thresholds from other trees
            tree_thresholds = [tree.get_thresholds(lb, ub) for tree in trees]

            # compute loss for the current node
            s = sum(
                [w * ss for w, (ss, _) in zip(tree_weights, tree_thresholds) if w > 0],
                np.zeros(self._my_state.n_out)
            )
            s += y.sum(0)
            w = s.sum()  # this assumes this is a classifier
            if d == 0 and w == 0:
                # special case: our previous fit and all of our neighbors have the
                # same, zero-length domain in the imput data
                log.warning('Domain of all trees and input data is length 0. '
                            'Copying an arbitrary tree. '
                            'Trees include any neighbors and any previous fit.\n'
                            'Trees: %s\nTrain target: %s',
                            trees, y)
                some_tree = next(iter(trees))
                self._my_state = copy.deepcopy(some_tree)
                return 0

            assert w > 0, (
                'BUG: a split with weight zero will cause infinite recursion.'
                f'w: {w}\ns: {s}\ny: {y}\n'
                f'tree_thresholds: {tree_thresholds}\n'
                f'tree_weights: {tree_weights}\n'
            )
            log.debug('weight of this subtree: %s', w)
            v = s / w
            v_l: float = self._loss_fn(v)
            vol = np.prod(ub - lb)

            # if at depth zero, set the total weight of samples and other trees
            if d == 0:
                n = w

            # check stopping conditions
            best_feat = -1  # used as a flag for whether a suitable split was found
            best_thres = None
            best_equal = None
            if (self._max_depth is None or d < self._max_depth) \
                    and w >= 2 * self._min_sample_weight \
                    and vol >= 2 * self._min_vol \
                    and v_l > 0:
                log.debug('splitting')
                # loop over features to evaluate candidate splits
                best_loss = v_l - self._ccp_alpha * n / w
                # permute to break ties randomly
                for f in self._random_state.permutation(self._my_state.n_in):
                    # compile candidate splits from data and other trees into arrays
                    t_list = []
                    eq_list = []
                    ddsdt_list = []
                    for weight, (ss, thres) in zip(tree_weights, tree_thresholds):
                        if weight > 0:
                            t_list += thres.ts[f]
                            eq_list += thres.eqs[f]
                            ddsdt_list += [weight * ddsdt for ddsdt in thres.ddsdts[f]]
                    t = np.concatenate((x[:, f], x[:, f], np.array(t_list)))
                    eq = np.concatenate((np.zeros(x.shape[0], dtype=bool),
                                         np.ones(x.shape[0], dtype=bool),
                                         np.array(eq_list)))
                    ds = np.zeros((t.shape[0], self._my_state.n_out))
                    if y.shape[0] > 0:
                        ds[y.shape[0]:2 * y.shape[0]] = y
                    ddsdt = np.zeros((t.shape[0], self._my_state.n_out))
                    if len(ddsdt_list) > 0:
                        ddsdt[-len(ddsdt_list):] = np.array(ddsdt_list)

                    # sort the split information by ascending threshold value
                    idx: np.ndarray = np.lexsort((eq, t))
                    t = t[idx]
                    eq = eq[idx]
                    ds = ds[idx]
                    ddsdt = ddsdt[idx]

                    # compute s for the left child at each candidate split via running sums
                    dt = t[1:] - t[:-1]
                    dsdt = np.cumsum(ddsdt, axis=0)
                    ds[1:] += dsdt[:-1] * dt[:, None]
                    s_l = np.cumsum(ds, axis=0)

                    # Correct for numerical error introduced by a long cumulative sum
                    # s_l[0] should be zero, s_l[-1] should be the same as s
                    # scale the valeus in s_l to ensure this is always correct.
                    s_l *= (s / np.where(s_l[-1] == 0, 1, s_l[-1]))[None, :]

                    # compute the loss for each candidate split
                    # TODO(Merritt): pair w Jack to implement solution for weights being 0 from KDDT
                    w_l: np.ndarray = s_l.sum(axis=1)
                    # ^ assumes this is a classifier, change to implement regression
                    v_l_: np.ndarray = s_l / (w_l[:, None] + 1e-12)
                    s_r: np.ndarray = s - s_l
                    w_r: np.ndarray = w - w_l
                    v_r: np.ndarray = s_r / (w_r[:, None] + 1e-12)

                    len_along_f: np.ndarray = ub[f] - lb[f]
                    vol_l: Union[int, np.ndarray] = (
                        0 if len_along_f == 0
                        else vol * (t - lb[f]) / len_along_f
                    )
                    vol_r: np.ndarray = vol - vol_l
                    loss: np.ndarray = (
                        w_l * self._loss_fn(v_l_) + w_r * self._loss_fn(v_r)
                    ) / (w + 1e-12)  # HACK TODO

                    # mask out splits that violate stopping conditions
                    loss[w_l < self._min_sample_weight] = np.inf
                    loss[w_r < self._min_sample_weight] = np.inf
                    loss[vol_l < self._min_vol] = np.inf
                    loss[vol_r < self._min_vol] = np.inf

                    # if weight of one split is 0, we would run forever.
                    loss[w_l == 0] = np.inf
                    loss[w_r == 0] = np.inf

                    # first and last element are splits where one child is identical to its parent,
                    # and should never be chosen.
                    loss[0] = np.inf
                    loss[-1] = np.inf

                    # avoid trying to split two data points with same value for f
                    # if threshold is lt and previous threshold is same it needs to be masked
                    loss[
                        np.logical_and(
                            np.logical_not(eq),
                            t == np.roll(t, 1)
                        )
                    ] = np.inf
                    # if threshold is lte and next threshold is same it needs to be masked
                    loss[
                        np.logical_and(
                            eq,
                            t == np.roll(t, -1)
                        )
                    ] = np.inf

                    # find the split with the best loss
                    # i and j are the lower and upper bounds of a range of
                    #   equivalent best candidate thresholds
                    i: int = int(loss.argmin())
                    j: int = i

                    # center splits in regions with (nearly) no change in label weights
                    eps = 1e-5  # a tolerance to accommodate numerical imprecision
                    while i > 0 and ds[i].sum() < eps:
                        i -= 1
                    while j < ds.shape[0] - 1 and ds[j + 1].sum() < eps:
                        j += 1

                    # update best split
                    if loss[i] < best_loss:
                        best_loss = loss[i]
                        best_feat = f
                        best_thres = t[i] if j == i else (t[i] + t[j]) / 2
                        best_equal = eq[i] if j == i else False
                        best_split_weight_lower = (w_l[i], w_r[i])
                        best_split_weight_upper = (w_l[j], w_r[j])
                        log.debug('best split on %s weight lower: %s', f, best_split_weight_lower)
                        log.debug('best split on %s weight upper: %s', f, best_split_weight_upper)
            else:
                log.debug('not splitting')

            # assign values to this node (as a leaf)
            i = len(self._my_state.children_left)
            self._append_leaf(v, w)

            # if a good split was found, perform the split and make this node internal
            if best_feat >= 0:
                assert best_thres is not None and best_equal is not None
                # grow left subtree
                mask = x[:, best_feat] <= best_thres if best_equal else x[:, best_feat] < best_thres
                old = ub[best_feat]
                ub[best_feat] = best_thres
                self._my_state.children_left[i] = recursive_fit(d + 1, x[mask], y[mask], lb, ub, n)
                ub[best_feat] = old

                # grow right subtree
                mask = np.logical_not(mask)
                old = lb[best_feat]
                lb[best_feat] = best_thres
                self._my_state.children_right[i] = recursive_fit(d + 1, x[mask], y[mask], lb, ub, n)
                lb[best_feat] = old

                # set this node's split parameters
                self._my_state.feature[i] = best_feat
                self._my_state.threshold[i] = best_thres
                self._my_state.equal[i] = best_equal

            # return this node's index so its parent can point to it
            return i

        _ = recursive_fit(0, x, y, self._my_state.lower_bound, self._my_state.upper_bound, 0)
        assert self.my_state is not None and len(self.my_state.feature) > 0, (
            'BUG: ended fit with an empty tree. '
            f'Input data: \n {dataset}'
        )

    def _predict_proba(self, x: np.ndarray) -> np.ndarray:
        def forward(i: int, x: np.ndarray) -> np.ndarray:
            log.debug('calling forward on node %s', i)
            assert self._predict_state is not None
            if self._is_leaf(i):
                return np.ones(x.shape[0])[:, None] * self._predict_state.value[i][None, :]
            (f, t, eq, left, right) = self._internal_node_info(i)
            out = np.zeros((x.shape[0], self._predict_state.n_out))
            assert f < x.shape[1], (
                f'BUG: feature {f} is out of bounds, x only has {x.shape[1]} features.  x:\n{x}'
            )
            mask = x[:, f] <= t if eq else x[:, f] < t
            out[mask] = forward(left, x[mask])
            mask = np.logical_not(mask)
            out[mask] = forward(right, x[mask])
            return out
        return forward(0, x)

    def _predict(self, dataset: Optional[Dataset], **kwargs) -> Optional[Dataset]:
        '''Predict the labels of the data. This is the actual implementation of predict.'''
        assert self._predict_state is not None, (
            'BUG: self._predict_state should not be None when predict is called.'
        )
        assert isinstance(self._predict_state, AutonDecisionTreeNeighbor), (
            'BUG: expected self._predict_state to be an AutonDecisionTreeNeighbor, '
            f'instead found {self._predict_state} of type {type(self._predict_state)}'
        )

        if dataset is None:
            log.error('Predict called on no data.')
            return None

        if DatasetKeys.DATAFRAME.value in dataset:
            x = dataset.dataframe.to_numpy()
        elif DatasetKeys.COVARIATES.value in dataset:
            x = dataset[DatasetKeys.COVARIATES.value].to_numpy()
        else:
            log.error('Predict called on empty or malformed dataset: %s',
                      dataset)
            return None

        assert x.shape[1] == self._predict_state.n_in, (
            f'BUG: Data given to predict has {x.shape[1]} columns but n_in is '
            f'{self._predict_state.n_in}.  Input data: \n{dataset}'
        )

        proba = self._predict_proba(x)
        # shape should be (x_nrow, n_out)
        assert proba.shape == (x.shape[0], self._predict_state.n_out)

        # shape should be (x_nrow, 1)
        # value for each row is the index of its prediction in the class list.
        predictions_class_id = np.argmax(proba, axis=1)

        predictions_class_name = [
            self._predict_state.class_list[i] for i in predictions_class_id
        ]

        target_col = dataset.metadata.target
        assert target_col is not None, 'must have target'

        proba_df = pd.DataFrame(proba, columns=self._predict_state.class_list)
        pred_df = pd.DataFrame({target_col.name: predictions_class_name})

        retval = dataset.output()

        retval.predictions = pred_df
        retval.probabilities = TableFactory(proba_df)  # type: ignore[abstract] # pylint: disable=abstract-class-instantiated
        return retval

    def norm2_diff(self, other: 'AutonDecisionTreeInstance') -> float:
        '''Compute the squared norm of the difference between two models.'''
        if self._predict_state is None or other.my_state is None:
            raise ValueError('Model must be trained before computing norm2_diff.')
        return self._predict_state.norm2_diff(other.my_state)

    @property
    def my_state(self) -> Optional[AutonDecisionTreeNeighbor]:
        '''Get the state of the model.

        Needed for type reconciliation.
        '''
        retval = super().my_state
        assert retval is None or isinstance(retval, AutonDecisionTreeNeighbor)
        return retval

    @property
    def my_state_copy(self) -> Optional[AutonDecisionTreeNeighbor]:
        '''Safely get a copy the state of the model.

        Needed for type reconciliation.
        '''
        retval = super().my_state_copy
        assert retval is None or isinstance(retval, AutonDecisionTreeNeighbor)
        return retval


class AutonDecisionTreeModel(Algorithm):
    '''Implementation of a decision tree classifier for distributed data.'''
    _name = 'auton_decision_tree_model'
    _instance_constructor = AutonDecisionTreeInstance
    _tags: Dict[str, List[str]] = {
        'task': [TaskType.MULTICLASS_CLASSIFICATION.name, TaskType.BINARY_CLASSIFICATION.name],
        'data_type': [DataType.TABULAR.name],
        'source': ['autonlab'],
        'distributed': ['true'],
        'supports_random_seed': ['true'],
    }
    _default_hyperparams = {
        'Lambda': 10.0,
        'omega': 2.0 / 3.0,
        'random_seed': Defaults.SEED,
        'tol': 0.000000001,
        'maxiter': None,
        'max_depth': None,
        'min_sample_weight': 0.0,
        # minimum number of weighted samples in an area needed to split
        'min_vol': 0.0,
        # Smallest volume of hyperrectangle within input space before we no longer split
        'ccp_alpha': 0.0,
        # cost complexity pruning: if gain in discrimination power is too small, we stop
        'loss_fn': loss_gini
    }


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = AutonDecisionTreeModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Tests for decider_auto.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pytest

from ...config_components.distributed_config import DistributedConfig
from ...wrangler.constants import PACKAGE_NAME

from ..stub_decider import DeciderStub
from .decider import Decider
from .decider_auto import DeciderCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code

FAKE_DECIDER_SOURCE = '''
from ngautonml.algorithms.impl.distributed_algorithm_instance import NeighborState
from ngautonml.config_components.distributed_config import DecidersConfig, DeciderConfig
from ngautonml.deciders.impl.decider import Decider
from ngautonml.deciders.impl.decider_catalog import DeciderCatalog

class FakeDeciderConfig(DeciderConfig):
    'Configuration for FakeDecider.'

class FakeDecider(Decider):
    name = 'fake_decider'
    tags = {
        'some_tag': ['some_value']
    }

    def decide(self, my_state: NeighborState, neighbor: NeighborState) -> bool:
        return False


def register(catalog: DeciderCatalog):
    catalog.register(FakeDecider)
    DecidersConfig.register(name=FakeDecider.name, config_type=FakeDeciderConfig)
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = (retval / f'.{PACKAGE_NAME}' / 'plugins'
                   / 'deciders' / 'fake_decider.py')
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_DECIDER_SOURCE)
    return retval


def test_decider_catalog_auto() -> None:
    dut = DeciderCatalogAuto()

    distributed = DistributedConfig(
        clause={
            'decider': {
                'stub_decider': {}
            },
        }
    )

    got = dut.lookup_by_name('stub_decider')(
        config=distributed.decider.stub_decider  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
    )
    assert isinstance(got, DeciderStub)


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)

    distributed = DistributedConfig(
        clause={
            'decider': {
                'fake_decider': {}
            },
        }
    )

    dut = DeciderCatalogAuto()
    got = dut.lookup_by_name('fake_decider')(
        config=distributed.decider.fake_decider)  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
    assert isinstance(got, Decider)
    assert got.tags['some_tag'] == ['some_value']



'''Tests for auton_linear_regression.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code,too-many-locals

from copy import deepcopy
import pickle
import socket
from typing import Tuple

import numpy as np
import pandas as pd
import pytest
from sklearn import datasets  # type: ignore[import]

from ...algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ...algorithms.impl.synchronous import advance
from ...config_components.distributed_config import DistributedConfig
from ...conftest import Clearer, Waiter
from ...metrics.impl.metric_auto import MetricCatalogAuto
from ...wrangler.base_port import BasePort  # type: ignore[import]
from ...wrangler.dataset import Column, Dataset, Metadata, RoleName
from .auton_linear_regression import (
    AutonLinearRegression,
    AutonLinearRegressionInstance,
    AutonLinearRegressionNeighbor)

from ...wrangler.logger import Logger
logger = Logger(__file__, to_stdout=False).logger()


CORRECT_V_L2 = 10
CORRECT_V = np.array([
    19.80426349, -1.23138257, 71.81414651,
    52.85306555, 19.89796648, 14.13228464, -45.26142198,
    47.34059162, 67.39734652, 43.43461662, 153.30386624])
CORRECT_V_MSE = 4436.62
GOOD_ENOUGH_MSE = 4300.0
BAD_MSE = 5000
CORRECT_DIST_PARAMS = {
    'L2': CORRECT_V_L2,
    'Lambda': 100000000,
    'omega': .9
}


def load_regression_dataset() -> Tuple[Dataset, Dataset, Dataset, Dataset]:
    # Load the diabetes dataset
    diabetes_x, diabetes_y = datasets.load_diabetes(return_X_y=True)

    # Use only one feature
    # diabetes_x = diabetes_x[:, np.newaxis, 2]

    # Split the data into training/testing sets
    diabetes_x_train = diabetes_x[:-20]
    diabetes_x_test = diabetes_x[-20:]

    # Split the targets into training/testing sets
    diabetes_y_train = diabetes_y[:-20]
    diabetes_y_test = diabetes_y[-20:]
    columns = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
    metadata = Metadata(roles={RoleName.TARGET: [Column('target')]})
    dataset_train = Dataset(
        metadata=metadata,
        covariates=pd.DataFrame(diabetes_x_train, columns=columns),
        target=pd.DataFrame({'target': diabetes_y_train})
    )
    dataset_test = Dataset(
        metadata=metadata,
        covariates=pd.DataFrame(diabetes_x_test, columns=columns)
    )
    ground_truth = Dataset(
        metadata=metadata,
        ground_truth=pd.DataFrame({'target': diabetes_y_test})
    )

    reduced_train = Dataset(
        covariates=pd.DataFrame(np.zeros((2, len(columns))), columns=columns),
        target=pd.DataFrame({'target': diabetes_y_train[:2]}),
        metadata=metadata
    )
    return (dataset_train, dataset_test, ground_truth, reduced_train)


base_port = BasePort()

LOOPBACK = '127.0.0.1'
SUNNY_DAY_PORT = base_port.next()


def test_sunny_day(assert_no_exceptions: Clearer) -> None:
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '2': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('2', (LOOPBACK, SUNNY_DAY_PORT)),
                ]
            },
        },
        'my_id': 2,
    })

    sklearn_alg = AlgorithmCatalogAuto().lookup_by_name('sklearn.linear_model.Ridge')

    alg = AutonLinearRegression()

    # Setting Lambda and L2 values to make our results similar to
    #   sklearn's logistic regression, in order to be more certain
    #   that our algorithm is reasonable.
    dut = alg.instantiate(
        distributed=distributed_config,
        Lambda=0.0, L2=CORRECT_V_L2)
    assert isinstance(dut, AutonLinearRegressionInstance)
    sklearn_dut = sklearn_alg.instantiate(
        alpha=CORRECT_V_L2
    )

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    train, test, ground_truth, _ = load_regression_dataset()

    try:
        dut.start()
        dut.fit(train)
        sklearn_dut.fit(train)

        sklearn_dut_v = np.array(list(sklearn_dut.impl.coef_) + [sklearn_dut.impl.intercept_])
        dut_v = dut.params

        result = dut.predict(test)
        sklearn_result = sklearn_dut.predict(test)

        sklearn_mse = metric.calculate(pred=sklearn_result,
                                       ground_truth=ground_truth
                                       )

        result_mse = metric.calculate(pred=result,
                                      ground_truth=ground_truth
                                      )

        assert dut_v == pytest.approx(sklearn_dut_v, abs=0.01)
        assert result_mse == pytest.approx(expected=sklearn_mse, abs=0.01)
        assert result_mse == pytest.approx(CORRECT_V_MSE, abs=0.1)
        assert_no_exceptions([dut])
    finally:
        dut.stop()


SELF_CONVERGENCE_PORT = base_port.next()


# TODO(Merritt): fix this
@pytest.mark.skip(
    reason='This does not currently work as linear regression has no self-regularization '
    'with no neighbors.  It should be reworked to use 2 nodes, or removed.')
def test_self_convergence(wait_til_all_fit: Waiter) -> None:
    '''Refit if our own state changes, even if we don't see a message from a neighbor.

    Scenario: there is one node with high self-regularization (low omega) that fits on dataset A.
    Then it additionally fits on dataset B, which changes the model enough to send.
    Fitting a third time on dataset B would once again change the model enough to send, due
    to the high self-regularization priotitizing the training from dataset A.
    We would like to continue refitting on dataset B until the model converges with itself.
    '''

    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '2': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('2', (LOOPBACK, SELF_CONVERGENCE_PORT)),
                ]
            },
        },
        'my_id': 2,
    })

    alg = AutonLinearRegression()

    dut = alg.instantiate(
        distributed=distributed_config,
        Lambda=1.0, L2=CORRECT_V_L2, omega=0.1)
    assert isinstance(dut, AutonLinearRegressionInstance)

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    train, test, ground_truth, reduced_train = load_regression_dataset()

    try:
        dut.start()
        dut.fit(reduced_train)

        wait_til_all_fit([dut])

        unconverged_result = dut.predict(test)

        unconverged_result_mse = metric.calculate(
            pred=unconverged_result,
            ground_truth=ground_truth
        )

        assert unconverged_result_mse != pytest.approx(
            expected=CORRECT_V_MSE, abs=100)

        dut.fit(train)

        wait_til_all_fit([dut], convergence_check=True)

        converged_result = dut.predict(test)

        conv_result_mse = metric.calculate(
            pred=converged_result,
            ground_truth=ground_truth
        )

        assert conv_result_mse == pytest.approx(CORRECT_V_MSE, abs=0.1)
    finally:
        dut.stop()


RECEIVE_EVENT_SENDER_PORT = base_port.next()
RECEIVE_EVENT_RECEIVER_PORT = base_port.next()


def test_receive_event(assert_no_exceptions: Clearer) -> None:
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                },
            },
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, RECEIVE_EVENT_SENDER_PORT)),
                    ('2', (LOOPBACK, RECEIVE_EVENT_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 2,
    })

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    _, test, ground_truth, reduced_train = load_regression_dataset()

    dut = AutonLinearRegression().instantiate(
        distributed=distributed_config, synchronous=True, **CORRECT_DIST_PARAMS)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, RECEIVE_EVENT_SENDER_PORT))

    # These parameters were hand-extracted from a successful fit.
    message = pickle.dumps((
        CORRECT_V, None
    ))

    try:
        dut.start()
        dut.fit(reduced_train)
        advance([dut])
        got_reduced = dut.predict(test)

        reduced_metric = metric.calculate(pred=got_reduced,
                                          ground_truth=ground_truth
                                          )
        assert reduced_metric > BAD_MSE

        sock.sendto(message, (LOOPBACK, RECEIVE_EVENT_RECEIVER_PORT))

        # Wait for the message to arrive and get processed.
        for _ in range(5):
            advance([dut])

        # Confirm that we learned from our neighbor.
        got = dut.predict(test)

        assert metric.calculate(pred=got,
                                ground_truth=ground_truth
                                ) < reduced_metric - 50
        assert_no_exceptions([dut])
    finally:
        dut.stop()
        sock.close()


SEND_SENDER_PORT = base_port.next()
SEND_RECEIVER_PORT = base_port.next()


def test_send(assert_no_exceptions: Clearer) -> None:
    distributed_config = DistributedConfig(clause={
        'polling_interval': '1.0',  # formerly 0.1
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, SEND_SENDER_PORT)),
                    ('2', (LOOPBACK, SEND_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 1,
    })

    train, _, _, _ = load_regression_dataset()

    receiver = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    receiver.bind((LOOPBACK, SEND_RECEIVER_PORT))

    dut = None
    try:
        dut = AutonLinearRegression().instantiate(distributed=distributed_config,
                                                  **CORRECT_DIST_PARAMS)
        dut.start()
        dut.fit(train)

        got_raw = receiver.recv(1024)  # Wait for the message to arrive and get processed.
        got = AutonLinearRegressionNeighbor.decode(got_raw)

        assert got.v == pytest.approx(CORRECT_V, abs=0.001)
        assert_no_exceptions([dut])

    finally:
        receiver.close()
        if dut is not None:
            dut.stop()


INTEGRATED_RECEIVER_PORT = base_port.next()
INTEGRATED_SENDER_PORT = base_port.next()


def test_integrated(assert_no_exceptions: Clearer) -> None:
    distributed_clause = {
        'polling_interval': '0.1',
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, INTEGRATED_RECEIVER_PORT)),
                    ('2', (LOOPBACK, INTEGRATED_SENDER_PORT)),
                ]
            }
        },
        'my_id': 0,
        'fit_eps': .01
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    train, test, ground_truth, reduced_train = load_regression_dataset()

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    alg = AutonLinearRegression()
    bad_data_node = alg.instantiate(distributed=DistributedConfig(distributed_receiver_clause),
                                    synchronous=True,
                                    **CORRECT_DIST_PARAMS)

    good_data_node = None
    try:
        bad_data_node.start()
        bad_data_node.fit(reduced_train)

        advance([bad_data_node])

        check = bad_data_node.predict(test)

        # Confirm that the training is very bad.
        assert metric.calculate(pred=check,
                                ground_truth=ground_truth
                                ) > BAD_MSE

        good_data_node = alg.instantiate(distributed=DistributedConfig(distributed_sender_clause),
                                         synchronous=True,
                                         **CORRECT_DIST_PARAMS)
        good_data_node.start()
        good_data_node.fit(train)

        # Wait for the message to arrive and get processed.
        advance([good_data_node, bad_data_node])

        # Confirm that we learned from our neighbor.
        got_receiver = bad_data_node.predict(test)
        got_sender = good_data_node.predict(test)

        assert metric.calculate(pred=got_receiver,
                                ground_truth=ground_truth
                                ) < BAD_MSE, 'receiver MSE not good enough'
        assert metric.calculate(pred=got_sender,
                                ground_truth=ground_truth
                                ) < BAD_MSE, 'sender MSE not good enough'
        assert_no_exceptions([good_data_node, bad_data_node])
    finally:
        bad_data_node.stop()
        if good_data_node is not None:
            good_data_node.stop()


NO_FIT_RECEIVER_PORT = base_port.next()
NO_FIT_SENDER_PORT = base_port.next()


def test_receive_no_fit(wait_til_all_fit: Waiter,
                        assert_no_exceptions: Clearer) -> None:
    '''Test that we spontaneously fit on a message from a neighbor without having seen data.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, NO_FIT_SENDER_PORT)),
                    ('2', (LOOPBACK, NO_FIT_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 2,
    })

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    _, test, ground_truth, _ = load_regression_dataset()

    dist_params = dict(CORRECT_DIST_PARAMS)
    dist_params['omega'] = 1.0  # zeros out self-regularization term

    dut = AutonLinearRegression().instantiate(distributed=distributed_config,
                                              **dist_params)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    try:
        sock.bind((LOOPBACK, NO_FIT_SENDER_PORT))
    except OSError as err:
        raise OSError(f'{LOOPBACK}:{NO_FIT_SENDER_PORT}') from err

    # These parameters were hand-extracted from a successful fit.
    message = pickle.dumps((
        CORRECT_V, None
    ))

    try:
        dut.start()
        sock.sendto(message, (LOOPBACK, NO_FIT_RECEIVER_PORT))

        # Wait for the message to arrive and get processed.
        wait_til_all_fit([dut])

        # Confirm that we learned from our neighbor.
        got = dut.predict(test)

        assert metric.calculate(pred=got,
                                ground_truth=ground_truth
                                ) == pytest.approx(expected=CORRECT_V_MSE, abs=0.1)
        assert_no_exceptions([dut])
    finally:
        dut.stop()
        sock.close()


TWO_HALF_RECEIVER_PORT = base_port.next()
TWO_HALF_SENDER_PORT = base_port.next()


def test_two_half() -> None:
    '''Two nodes with half the training behaves as (reasonably) expected.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, TWO_HALF_RECEIVER_PORT)),
                    ('2', (LOOPBACK, TWO_HALF_SENDER_PORT)),
                ]
            }
        },
        'my_id': 0,
        'fit_eps': .001
    }

    train, test, ground_truth, _ = load_regression_dataset()

    half_n = int(train['covariates'].shape[0] / 2)
    train_first = Dataset(
        metadata=train.metadata,
        covariates=train['covariates'].iloc[0:half_n],
        target=train['target'].iloc[0:half_n]
    )
    train_second = Dataset(
        metadata=train.metadata,
        covariates=train['covariates'].iloc[half_n:],
        target=train['target'].iloc[half_n:]
    )

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    dist_params_first = dict(CORRECT_DIST_PARAMS)
    dist_params_first['Lambda'] = 100000
    dist_params_second = dict(dist_params_first)

    alg = AutonLinearRegression()
    dut_first = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        synchronous=True,
        **dist_params_first)

    dut_second = alg.instantiate(
        distributed=DistributedConfig(distributed_sender_clause),
        synchronous=True,
        **dist_params_second)

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    try:
        dut_first.start()
        dut_second.start()

        dut_first.fit(train_first)
        dut_second.fit(train_second)

        # Wait for the message to arrive and get processed.
        for _ in range(20):
            advance([dut_first, dut_second])

        # Confirm that neighbor optimization doesn't mess up result.
        got_receiver = dut_first.predict(test)
        got_sender = dut_second.predict(test)

        assert dut_second.params == pytest.approx(dut_first.params, rel=.1), (
            f'params mismatch: {dut_second.params} vs {dut_first.params}'
        )

        assert metric.calculate(pred=got_receiver,
                                ground_truth=ground_truth
                                ) >= GOOD_ENOUGH_MSE, (
            'receiver MSE mismatch: '
            f'{metric.calculate(pred=got_receiver, ground_truth=ground_truth)}')

        assert metric.calculate(pred=got_sender,
                                ground_truth=ground_truth
                                ) >= GOOD_ENOUGH_MSE, (
            'sender MSE mismatch: '
            f'{metric.calculate(pred=got_sender, ground_truth=ground_truth)}')

    finally:
        dut_first.stop()
        if dut_second is not None:
            dut_second.stop()


TWO_SAME_RECEIVER_PORT = base_port.next()
TWO_SAME_SENDER_PORT = base_port.next()


def test_two_same(assert_no_exceptions: Clearer) -> None:
    '''Two nodes with the same training don't change their coefficients.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, TWO_SAME_RECEIVER_PORT)),
                    ('2', (LOOPBACK, TWO_SAME_SENDER_PORT)),
                ]
            }
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    train, test, ground_truth, _ = load_regression_dataset()

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    alg = AutonLinearRegression()
    dut_receiver = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        synchronous=True,
        **CORRECT_DIST_PARAMS)

    dut_sender = alg.instantiate(
        distributed=DistributedConfig(distributed_sender_clause),
        synchronous=True,
        **CORRECT_DIST_PARAMS)
    try:

        dut_receiver.start()
        dut_sender.start()

        dut_receiver.fit(train)
        dut_sender.fit(train)

        # Wait for the message to arrive and get processed.
        advance([dut_receiver, dut_sender])

        # Confirm that neighbor optimization doesn't mess up result.
        got_receiver = dut_receiver.predict(test)
        got_sender = dut_sender.predict(test)

        assert metric.calculate(pred=got_receiver,
                                ground_truth=ground_truth
                                ) == pytest.approx(CORRECT_V_MSE, abs=0.1), (
            'receiver MSE mismatch')

        assert metric.calculate(pred=got_sender,
                                ground_truth=ground_truth
                                ) == pytest.approx(CORRECT_V_MSE, abs=0.1), (
            'sender MSE mismatch')

        assert got_receiver is not None
        assert got_sender is not None

        pd.testing.assert_frame_equal(got_receiver.predictions, got_sender.predictions)

        assert dut_sender.params == pytest.approx(dut_receiver.params, abs=0.001), (
            f'receiver params mismatch: {dut_sender.params} vs {dut_receiver.params}')
        assert dut_sender.params == pytest.approx(CORRECT_V, abs=0.001), (
            f'correct params mismatch: {dut_sender.params} vs {CORRECT_V}')
        assert_no_exceptions([dut_sender, dut_receiver])
    finally:
        dut_receiver.stop()
        if dut_sender is not None:
            dut_sender.stop()


TEST_N_SAME_N = 5
TEST_N_SAME_PORTS = [base_port.next() for _ in range(TEST_N_SAME_N)]


def test_n_same(assert_no_exceptions: Clearer) -> None:
    '''N nodes fully connected with the same training match the non-distributed result.'''
    n = TEST_N_SAME_N

    adjacency = {f'{k}': [j for j in range(1, n + 1) if j != k] for k in range(1, n + 1)}
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_N_SAME_PORTS[k - 1])) for k in range(1, n + 1)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    train, test, ground_truth, _ = load_regression_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonLinearRegression()

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc),
        synchronous=True,
        **CORRECT_DIST_PARAMS) for dc in distributed_clauses]
    try:
        for dut in duts:
            dut.start()

        for dut in duts:
            dut.fit(train)

        # Wait for the message to arrive and get processed.
        for _ in range(20):
            advance(duts)

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(test) for dut in duts]

        scores = np.array([metric.calculate(pred=pred, ground_truth=ground_truth)
                           for pred in got_predictions])
        want = np.array([CORRECT_V_MSE] * len(scores))
        np.testing.assert_allclose(scores, want, rtol=0.1)

        gp0 = got_predictions[0]
        assert gp0 is not None
        for i in range(1, n):
            gpi = got_predictions[i]
            assert gpi is not None
            np.testing.assert_allclose(gp0.predictions, gpi.predictions, rtol=0.1)

        for i in range(1, n):
            assert duts[0].params == pytest.approx(duts[i].params, rel=0.1)

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


TEST_TELEPHONE_N = 5
TEST_TELEPHONE_PORTS = [base_port.next() for _ in range(TEST_TELEPHONE_N)]


def test_telephone(assert_no_exceptions: Clearer) -> None:
    '''Whisper down the lane test. Nodes connected in a line.
    The only node that has data is one at the end '''
    n = TEST_TELEPHONE_N

    adjacency = {f'{k}': [k - 1, k + 1] for k in range(2, n)}
    adjacency['1'] = [2]
    adjacency[f'{n}'] = [n - 1]
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_TELEPHONE_PORTS[k - 1]))
                           for k in range(1, n + 1)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('mean_squared_error')

    train, test, ground_truth, _ = load_regression_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonLinearRegression()

    telephone_params = dict(CORRECT_DIST_PARAMS, Lambda=1)
    telephone_params['omega'] = .5

    params = [dict(telephone_params) if k == 0 else dict(telephone_params) for k in range(n)]

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc), synchronous=True,
        **ps) for (dc, ps) in zip(distributed_clauses, params)]
    try:
        for dut in duts:
            dut.start()

        duts[0].fit(train)

        # Wait for the message to arrive and get processed.
        for _ in range(10 * TEST_TELEPHONE_N):
            advance(duts)

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(test) for dut in duts]

        for got_prediction in got_predictions:
            assert metric.calculate(pred=got_prediction,
                                    ground_truth=ground_truth
                                    ) == pytest.approx(CORRECT_V_MSE, abs=1), (
                f'got {metric.calculate(pred=got_prediction, ground_truth=ground_truth)}, '
                f'not {CORRECT_V_MSE}')

        gp0 = got_predictions[0]
        assert gp0 is not None, 'No predictions from node 0'
        for i in range(1, n):
            gpi = got_predictions[i]
            assert gpi is not None, f'No predictions from node {i}'
            np.testing.assert_allclose(gp0.predictions, gpi.predictions, rtol=.01)

        for i in range(n - 1):
            assert duts[i].params == pytest.approx(duts[i + 1].params, rel=0.01), (
                f'Node {i} and {i + 1} have different parameters')

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


COLS_SENDER_PORT = base_port.next()
COLS_RECEIVER_PORT = base_port.next()


def test_columns(assert_no_exceptions: Clearer) -> None:
    '''Test that columns are handled in data loading, sending and receiving.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                },
            },
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, COLS_SENDER_PORT)),
                    ('2', (LOOPBACK, COLS_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 2,
    })

    _, _, _, reduced_train = load_regression_dataset()

    dut = AutonLinearRegression().instantiate(distributed=distributed_config,
                                              synchronous=True,
                                              **CORRECT_DIST_PARAMS)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, COLS_SENDER_PORT))

    # These parameters were hand-extracted from a successful fit.
    match_message = pickle.dumps((
        CORRECT_V, reduced_train['covariates'].columns.to_list()
    ))

    mismatch_message = pickle.dumps((
        CORRECT_V, reduced_train['covariates'].columns.to_list() + ['hamster']
    ))

    try:
        dut.start()
        dut.fit(reduced_train)

        advance([dut])

        sock.sendto(match_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])
        assert_no_exceptions([dut])

        sock.sendto(mismatch_message, (LOOPBACK, COLS_RECEIVER_PORT))
        advance([dut])

        got = dut.poll_exceptions()
        assert got is not None
        assert isinstance(got[0], NotImplementedError)
        assert 'mismatch' in str(got[1]), f'Expected "mismatch" in {got[1]}'
        assert 'hamster' in str(got[1]), f'Expected "hamster" in {got[1]}'

        # make sure there are no other exceptions we don't expect
        assert_no_exceptions([dut])

    finally:
        dut.stop()
        sock.close()



'''Adds the Borda Rank Aggregator'''

# pylint: disable=too-many-locals

import importlib
from typing import Any
from typing_extensions import Protocol

import numpy as np

from ..instantiator.executable_pipeline import PipelineResults
from ..ranker.ranker import Rankings

from .impl.aggregator import Aggregator, AggregatorInsufficientPipelines, AggregatorRanking
from .impl.aggregator_catalog import AggregatorCatalog


class AggregatorImpl(Protocol):
    '''Type for TSAD aggregator implementations'''
    def __call__(self, ranks: np.ndarray, **kwargs: Any):
        ...


class TSADAggregator(Aggregator):
    '''Generic object for Aggregators from TSAD library.'''
    _aggregator: AggregatorImpl
    _high = False  # The best score is 1, and they get worse as you go up.

    def __init__(self, **params):
        super().__init__(**params)
        self._aggregator = self._load_module(f'.{self._name}')

    def _load_module(self, name: str):
        # Split name into module part (e.g. sklearn.linear_model)
        # and constructor part (e.g. LinearRegression)
        parts = name.split('.')
        constructor_part = parts[-1]
        module = importlib.import_module('.'.join(parts[:-1]), package='ngautonml.aggregators.impl')
        # Load the constructor.
        return getattr(module, constructor_part)

    def aggregate(self,
                  rankings: Rankings,
                  all_scores: bool = False, **overrides) -> AggregatorRanking:
        rankings_list = list(rankings.values())
        first_ranking = rankings_list[0]
        first_scores = first_ranking.scores(all_scores=all_scores)
        num_of_pipelines = len(first_scores)
        num_of_metrics = len(rankings_list)

        assert self._tags is not None, (
            'BUG: self._tags should have been initialized in TSADGenerator.__init__().'
        )
        if 'required_number_of_pipelines' in self._tags:
            num_of_pipelines_required = int(self._tags['required_number_of_pipelines'][0])
            if num_of_pipelines < num_of_pipelines_required:
                raise AggregatorInsufficientPipelines(
                    f'Given {num_of_pipelines} pipelines but '
                    f'require {num_of_pipelines_required} pipelines'
                )

        pipeline_to_index = {}
        index_to_pipeline = {}
        pipeline_results = PipelineResults()
        for i, scored_result in enumerate(first_scores):
            pipeline_des = scored_result.pipeline_des
            pipeline_to_index[pipeline_des] = i
            index_to_pipeline[i] = pipeline_des
            pipeline_results[pipeline_des] = scored_result.result

        # ranks[i, j] = the rank of the jth pipeline under the ith metric
        ranks = np.zeros((num_of_metrics, num_of_pipelines))
        for m, ranking in enumerate(rankings_list):
            for rank, scored_result in enumerate(ranking.scores(all_scores=all_scores)):
                pipeline_des = scored_result.pipeline_des
                n = pipeline_to_index[pipeline_des]
                ranks[m, n] = rank

        _, aggregate_rank = self._aggregator(ranks, **self.hyperparams(**overrides))

        scores = {}
        for i, rank in enumerate(aggregate_rank):
            scores[index_to_pipeline[rank]] = i

        return AggregatorRanking(metric=self, results=pipeline_results, scores=scores)


def register(catalog: AggregatorCatalog):  # pylint: disable=unused-argument
    '''Nothing to register.

    All subclasses of SklearnAlgorithm are registered in sklearn_algorithms.py
    '''
